from unittest.mock import Mock, patch

from click.testing import CliRunner
import pytest

from anyscale.client.openapi_client.models import ClusterManagementStackVersions
from anyscale.commands.cloud_commands import setup_cloud


@pytest.mark.parametrize(
    "use_legacy_stack", [True, False],
)
@pytest.mark.parametrize("enable_head_node_fault_tolerance", [True, False])
def test_cloud_setup(use_legacy_stack, enable_head_node_fault_tolerance):
    runner = CliRunner()
    mock_cloud_controller = Mock()
    mock_setup_managed_cloud = Mock()
    mock_cloud_controller().setup_managed_cloud = mock_setup_managed_cloud

    provider = "aws"
    region = "us-west-2"
    name = "unit-test-cloud"
    cluster_management_stack_version = (
        ClusterManagementStackVersions.V1
        if use_legacy_stack
        else ClusterManagementStackVersions.V2
    )

    args = ["--provider", provider, "--name", name, "--region", region]

    if use_legacy_stack:
        args.append("--use-legacy-stack")

    if enable_head_node_fault_tolerance:
        args.append("--enable-head-node-fault-tolerance")

    with patch(
        "anyscale.commands.cloud_commands.CloudController", new=mock_cloud_controller,
    ):
        runner.invoke(setup_cloud, args=args)

    mock_setup_managed_cloud.assert_called_once_with(
        provider=provider,
        region=region,
        name=name,
        functional_verify=None,
        cluster_management_stack_version=cluster_management_stack_version,
        enable_head_node_fault_tolerance=enable_head_node_fault_tolerance,
    )



# coding: utf-8

"""
    Anyscale API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from anyscale_client.configuration import Configuration


class SessionsQuery(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'name': 'TextQuery',
        'paging': 'PageQuery',
        'state_filter': 'list[SessionState]',
        'creator_id': 'str'
    }

    attribute_map = {
        'name': 'name',
        'paging': 'paging',
        'state_filter': 'state_filter',
        'creator_id': 'creator_id'
    }

    def __init__(self, name=None, paging=None, state_filter=[], creator_id=None, local_vars_configuration=None):  # noqa: E501
        """SessionsQuery - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._name = None
        self._paging = None
        self._state_filter = None
        self._creator_id = None
        self.discriminator = None

        if name is not None:
            self.name = name
        if paging is not None:
            self.paging = paging
        if state_filter is not None:
            self.state_filter = state_filter
        if creator_id is not None:
            self.creator_id = creator_id

    @property
    def name(self):
        """Gets the name of this SessionsQuery.  # noqa: E501

        Filters Cluster by name. If this field is absent, no filtering is done.  # noqa: E501

        :return: The name of this SessionsQuery.  # noqa: E501
        :rtype: TextQuery
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this SessionsQuery.

        Filters Cluster by name. If this field is absent, no filtering is done.  # noqa: E501

        :param name: The name of this SessionsQuery.  # noqa: E501
        :type: TextQuery
        """

        self._name = name

    @property
    def paging(self):
        """Gets the paging of this SessionsQuery.  # noqa: E501

        Pagination information.  # noqa: E501

        :return: The paging of this SessionsQuery.  # noqa: E501
        :rtype: PageQuery
        """
        return self._paging

    @paging.setter
    def paging(self, paging):
        """Sets the paging of this SessionsQuery.

        Pagination information.  # noqa: E501

        :param paging: The paging of this SessionsQuery.  # noqa: E501
        :type: PageQuery
        """

        self._paging = paging

    @property
    def state_filter(self):
        """Gets the state_filter of this SessionsQuery.  # noqa: E501

        Filter Cluster by Session State. If this field is an empty set, no filtering is done.  # noqa: E501

        :return: The state_filter of this SessionsQuery.  # noqa: E501
        :rtype: list[SessionState]
        """
        return self._state_filter

    @state_filter.setter
    def state_filter(self, state_filter):
        """Sets the state_filter of this SessionsQuery.

        Filter Cluster by Session State. If this field is an empty set, no filtering is done.  # noqa: E501

        :param state_filter: The state_filter of this SessionsQuery.  # noqa: E501
        :type: list[SessionState]
        """

        self._state_filter = state_filter

    @property
    def creator_id(self):
        """Gets the creator_id of this SessionsQuery.  # noqa: E501

        Filters Jobs by creator_id. If this field is absent, no filtering is done.  # noqa: E501

        :return: The creator_id of this SessionsQuery.  # noqa: E501
        :rtype: str
        """
        return self._creator_id

    @creator_id.setter
    def creator_id(self, creator_id):
        """Sets the creator_id of this SessionsQuery.

        Filters Jobs by creator_id. If this field is absent, no filtering is done.  # noqa: E501

        :param creator_id: The creator_id of this SessionsQuery.  # noqa: E501
        :type: str
        """

        self._creator_id = creator_id

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, SessionsQuery):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, SessionsQuery):
            return True

        return self.to_dict() != other.to_dict()



import re
from typing import List, Optional

import click

from anyscale.cli_logger import BlockLogger
from anyscale.client.openapi_client.models import ClusterManagementStackVersions
from anyscale.commands.util import OptionPromptNull
from anyscale.controllers.cloud_controller import CloudController


log = BlockLogger()  # CLI Logger


@click.group(
    "cloud",
    short_help="Configure cloud provider authentication for Anyscale.",
    help="""Configure cloud provider authentication and setup
to allow Anyscale to launch instances in your account.""",
)
def cloud_cli() -> None:
    pass


@cloud_cli.command(name="delete", help="Delete a cloud.")
@click.argument("cloud-name", required=False)
@click.option("--name", "-n", help="Delete cloud by name.", type=str)
@click.option(
    "--cloud-id",
    "--id",
    help="Cloud id to delete. Alternative to cloud name.",
    required=False,
)
@click.option(
    "--yes", "-y", is_flag=True, default=False, help="Don't ask for confirmation."
)
def cloud_delete(
    cloud_name: Optional[str], name: Optional[str], cloud_id: Optional[str], yes: bool
) -> None:
    if cloud_name and name and cloud_name != name:
        raise click.ClickException(
            "The positional argument CLOUD_NAME and the keyword argument --name "
            "were both provided. Please only provide one of these two arguments."
        )
    CloudController().delete_cloud(
        cloud_name=cloud_name or name, cloud_id=cloud_id, skip_confirmation=yes
    )


@cloud_cli.command(
    name="set-default",
    help=(
        "Sets default cloud for your organization. This operation can only be performed "
        "by organization admins, and the default cloud must have organization level "
        "permissions."
    ),
)
@click.argument("cloud-name", required=False)
@click.option("--name", "-n", help="Set cloud as default by name.", type=str)
@click.option(
    "--cloud-id",
    "--id",
    help="Cloud id to set as default. Alternative to cloud name.",
    required=False,
)
def cloud_set_default(
    cloud_name: Optional[str], name: Optional[str], cloud_id: Optional[str]
) -> None:
    if cloud_name and name and cloud_name != name:
        raise click.ClickException(
            "The positional argument CLOUD_NAME and the keyword argument --name "
            "were both provided. Please only provide one of these two arguments."
        )
    CloudController().set_default_cloud(
        cloud_name=cloud_name or name, cloud_id=cloud_id
    )


@cloud_cli.command(name="setup", help="Set up a cloud provider.")
@click.option(
    "--provider",
    help="The cloud provider type.",
    required=True,
    prompt="Provider",
    type=click.Choice(["aws", "gcp"], case_sensitive=False),
)
@click.option(
    "--region",
    cls=OptionPromptNull,
    help="Region to set up the credentials in.",
    required=True,
    prompt="Region",
    default_option="provider",
    default=lambda p: "us-west-2" if p == "aws" else "us-west1",
    show_default=True,
)
@click.option("--name", "-n", help="Name of the cloud.", required=True, prompt="Name")
@click.option(
    "--project-id",
    help="Globally Unique project ID for GCP clouds (e.g., my-project-abc123)",
    required=False,
    type=str,
)
@click.option(
    "--functional-verify",
    help="Verify the cloud is functional. This will check that the cloud can launch workspace/service.",
    required=False,
    is_flag=False,
    flag_value="workspace",
)
@click.option(
    "--anyscale-managed",
    is_flag=True,
    default=False,
    help="Let anyscale create all the resources.",
)
@click.option(
    "--use-legacy-stack",
    is_flag=True,
    default=False,
    hidden=True,
    help="Whether to use the legacy stack.",
)
@click.option(
    "--enable-head-node-fault-tolerance",
    is_flag=True,
    default=False,
    help="Whether to enable head node fault tolerance for services.",
)
def setup_cloud(  # noqa: PLR0913
    provider: str,
    region: str,
    name: str,
    project_id: str,
    functional_verify: Optional[str],
    anyscale_managed: bool,  # noqa: ARG001
    use_legacy_stack: bool,
    enable_head_node_fault_tolerance: bool,
) -> None:
    # TODO (congding): remove `anyscale_managed` in the future, now keeping it for compatibility
    cluster_management_stack_version = (
        ClusterManagementStackVersions.V1
        if use_legacy_stack
        else ClusterManagementStackVersions.V2
    )
    if provider == "aws":
        CloudController().setup_managed_cloud(
            provider=provider,
            region=region,
            name=name,
            functional_verify=functional_verify,
            cluster_management_stack_version=cluster_management_stack_version,
            enable_head_node_fault_tolerance=enable_head_node_fault_tolerance,
        )
    elif provider == "gcp":
        if not project_id:
            project_id = click.prompt("GCP Project ID", type=str)
        CloudController().setup_managed_cloud(
            provider=provider,
            region=region,
            name=name,
            project_id=project_id,
            functional_verify=functional_verify,
            cluster_management_stack_version=cluster_management_stack_version,
            enable_head_node_fault_tolerance=enable_head_node_fault_tolerance,
        )


@cloud_cli.command(
    name="list", help=("List information about clouds in your Anyscale organization."),
)
@click.option(
    "--name",
    "-n",
    required=False,
    default=None,
    help="Name of cloud to get information about.",
)
@click.option(
    "--cloud-id",
    "--id",
    required=False,
    default=None,
    help=("Id of cloud to get information about."),
)
def list_cloud(name: Optional[str], cloud_id: Optional[str],) -> None:
    print(CloudController().list_clouds(cloud_name=name, cloud_id=cloud_id,))


@cloud_cli.group("config", help="Manage the configuration for a cloud.")
def cloud_config_group() -> None:
    pass


@cloud_config_group.command("update", help="Update the configuration for a cloud.")
@click.argument("cloud-name", required=False)
@click.option(
    "--cloud-id",
    "--id",
    help="Cloud id to update. Alternative to cloud name.",
    required=False,
)
@click.option("--name", "-n", help="Update configuration of cloud by name.", type=str)
@click.option(
    "--max-stopped-instances",
    help="Maximum number of stopped instances permitted in the shared instance pool.",
    required=True,
)
def cloud_config_update(
    cloud_name: Optional[str],
    name: Optional[str],
    cloud_id: Optional[str],
    max_stopped_instances: int,
) -> None:
    if cloud_name and name and cloud_name != name:
        raise click.ClickException(
            "The positional argument CLOUD_NAME and the keyword argument --name "
            "were both provided. Please only provide one of these two arguments."
        )
    CloudController().update_cloud_config(
        cloud_name=cloud_name or name,
        cloud_id=cloud_id,
        max_stopped_instances=max_stopped_instances,
    )


@cloud_config_group.command("get", help="Get the current configuration for a cloud.")
@click.argument("cloud-name", required=False)
@click.option("--name", "-n", help="Update configuration of cloud by name.", type=str)
@click.option(
    "--cloud-id",
    "--id",
    help="Cloud id to get details about. Alternative to cloud name.",
    required=False,
)
def cloud_config_get(
    cloud_name: Optional[str], name: Optional[str], cloud_id: Optional[str]
) -> None:
    if cloud_name and name and cloud_name != name:
        raise click.ClickException(
            "The positional argument CLOUD_NAME and the keyword argument --name "
            "were both provided. Please only provide one of these two arguments."
        )
    print(
        CloudController().get_cloud_config(
            cloud_name=cloud_name or name, cloud_id=cloud_id,
        )
    )


@cloud_cli.command(
    name="register", help="Register an anyscale cloud with your own resources."
)
@click.option(
    "--provider",
    help="The cloud provider type.",
    required=True,
    type=click.Choice(["aws", "gcp"], case_sensitive=False),
)
@click.option(
    "--region",
    cls=OptionPromptNull,
    help="Region to set up the credentials in.",
    required=True,
    default_option="provider",
    default=lambda p: "us-west-2" if p == "aws" else "us-west1",
    show_default=True,
)
@click.option(
    "--name", "-n", help="Name of the cloud.", required=True,
)
@click.option(
    "--vpc-id", help="The ID of the VPC.", required=False, type=str,
)
@click.option(
    "--subnet-ids",
    help="Comma separated list of subnet ids.",
    required=False,
    type=str,
)
@click.option(
    "--efs-id", help="The EFS ID.", required=False, type=str,
)
@click.option(
    "--anyscale-iam-role-id",
    help="The Anyscale IAM Role ARN.",
    required=False,
    type=str,
)
@click.option(
    "--instance-iam-role-id",
    help="The instance IAM role ARN.",
    required=False,
    type=str,
)
@click.option(
    "--security-group-ids",
    help="IDs of the security groups.",
    required=False,
    type=str,
)
@click.option(
    "--s3-bucket-id", help="S3 bucket ID.", required=False, type=str,
)
@click.option(
    "--project-id",
    help="Globally Unique project ID for GCP clouds (e.g., my-project-abc123)",
    required=False,
    type=str,
)
@click.option(
    "--vpc-name", help="VPC name for GCP clouds", required=False, type=str,
)
@click.option(
    "--subnet-names",
    help="Comma separated list of subnet names for GCP clouds",
    required=False,
    type=str,
)
@click.option(
    "--filestore-instance-id",
    help="Filestore instance ID for GCP clouds.",
    required=False,
    type=str,
)
@click.option(
    "--filestore-location",
    help="Filestore location for GCP clouds.",
    required=False,
    type=str,
)
@click.option(
    "--anyscale-service-account-email",
    help="Anyscale service account email for GCP clouds.",
    required=False,
    type=str,
)
@click.option(
    "--instance-service-account-email",
    help="Instance service account email for GCP clouds.",
    required=False,
    type=str,
)
@click.option(
    "--provider-name",
    help="Workload Identity Federation provider name for Anyscale access.",
    required=False,
    type=str,
)
@click.option(
    "--firewall-policy-names",
    help="Filewall policy names for GCP clouds",
    required=False,
    type=str,
)
@click.option(
    "--cloud-storage-bucket-name",
    help="Cloud storage bucket name for GCP clouds",
    required=False,
    type=str,
)
@click.option(
    "--private-network", help="Use private network.", is_flag=True, default=False,
)
@click.option(
    "--use-legacy-stack",
    is_flag=True,
    default=False,
    hidden=True,
    help="Whether to use the legacy stack.",
)
@click.option(
    "--functional-verify",
    help="Verify the cloud is functional. This will check that the cloud can launch workspace/service.",
    required=False,
    is_flag=False,
    flag_value="workspace",
)
@click.option(
    "--yes", "-y", is_flag=True, default=False, help="Skip asking for confirmation."
)
def register_cloud(  # noqa: PLR0913
    provider: str,
    region: str,
    name: str,
    vpc_id: str,
    subnet_ids: str,
    efs_id: str,
    anyscale_iam_role_id: str,
    instance_iam_role_id: str,
    security_group_ids: str,
    s3_bucket_id: str,
    project_id: str,
    vpc_name: str,
    subnet_names: str,
    filestore_instance_id: str,
    filestore_location: str,
    anyscale_service_account_email: str,
    instance_service_account_email: str,
    provider_name: str,
    firewall_policy_names: str,
    cloud_storage_bucket_name: str,
    functional_verify: Optional[str],
    private_network: bool,
    use_legacy_stack: bool,
    yes: bool,
) -> None:
    missing_args: List[str] = []
    if provider == "aws":
        for resource in [
            (vpc_id, "--vpc-id"),
            (subnet_ids, "--subnet-ids"),
            (efs_id, "--efs-id"),
            (anyscale_iam_role_id, "--anyscale-iam-role-id"),
            (instance_iam_role_id, "--instance-iam-role-id"),
            (security_group_ids, "--security-group-ids"),
            (s3_bucket_id, "--s3-bucket-id"),
        ]:
            if resource[0] is None:
                missing_args.append(resource[1])

        if len(missing_args) > 0:
            raise click.ClickException(f"Please provide a value for {missing_args}")

        CloudController().register_aws_cloud(
            region=region,
            name=name,
            vpc_id=vpc_id,
            subnet_ids=subnet_ids.split(","),
            efs_id=efs_id,
            anyscale_iam_role_id=anyscale_iam_role_id,
            instance_iam_role_id=instance_iam_role_id,
            security_group_ids=security_group_ids.split(","),
            s3_bucket_id=s3_bucket_id,
            functional_verify=functional_verify,
            private_network=private_network,
            cluster_management_stack_version=ClusterManagementStackVersions.V1
            if use_legacy_stack
            else ClusterManagementStackVersions.V2,
            yes=yes,
        )
    elif provider == "gcp":
        # Keep the parameter naming ({resource}_name or {resource}_id) consistent with GCP to reduce confusion for customers
        for resource in [
            (project_id, "--project-id"),
            (vpc_name, "--vpc-name"),
            (subnet_names, "--subnet-names"),
            (filestore_instance_id, "--filestore-instance-id"),
            (filestore_location, "--filestore-location"),
            (anyscale_service_account_email, "--anyscale-service-account-email"),
            (instance_service_account_email, "--instance-service-account-email"),
            (provider_name, "--provider-name"),
            (firewall_policy_names, "--firewall-policy-names"),
            (cloud_storage_bucket_name, "--cloud-storage-bucket-name"),
        ]:
            if resource[0] is None:
                missing_args.append(resource[1])

        if len(missing_args) > 0:
            raise click.ClickException(f"Please provide a value for {missing_args}")

        if project_id[0].isdigit():
            # project ID should start with a letter
            raise click.ClickException(
                "Please provide a valid project ID. Note that project ID is not project number, see https://cloud.google.com/resource-manager/docs/creating-managing-projects#before_you_begin for details."
            )

        if (
            re.search(
                "projects/[0-9]*/locations/global/workloadIdentityPools/.+/providers/.+",
                provider_name,
            )
            is None
        ):
            raise click.ClickException(
                "Please provide a valid, fully qualified provider name. Example: projects/<project number>/locations/global/workloadIdentityPools/<pool name>/providers/<provider id>"
            )

        CloudController().register_gcp_cloud(
            region=region,
            name=name,
            project_id=project_id,
            vpc_name=vpc_name,
            subnet_names=subnet_names.split(","),
            filestore_instance_id=filestore_instance_id,
            filestore_location=filestore_location,
            anyscale_service_account_email=anyscale_service_account_email,
            instance_service_account_email=instance_service_account_email,
            # TODO (allenyin): use provider_name instead of provider_id everywhere.
            provider_id=provider_name,
            firewall_policy_names=firewall_policy_names.split(","),
            cloud_storage_bucket_name=cloud_storage_bucket_name,
            functional_verify=functional_verify,
            private_network=private_network,
            cluster_management_stack_version=ClusterManagementStackVersions.V1
            if use_legacy_stack
            else ClusterManagementStackVersions.V2,
            yes=yes,
        )
    else:
        raise click.ClickException(
            f"Invalid Cloud provider: {provider}. Available providers are [aws, gcp]."
        )


@cloud_cli.command(name="verify", help="Checks the healthiness of a cloud.")
@click.argument("cloud-name", required=False)
@click.option("--name", "-n", help="Verify cloud by name.", type=str)
@click.option(
    "--cloud-id",
    "--id",
    help="Verify cloud by cloud id, alternative to cloud name.",
    required=False,
)
@click.option(
    "--functional-verify",
    help="Verify the cloud is functional. This will check that the cloud can launch workspace/service.",
    required=False,
    is_flag=False,
    flag_value="workspace",
)
@click.option(
    "--strict",
    is_flag=True,
    default=False,
    help="Strict Verify. Treat warnings as failures.",
)
def cloud_verify(
    cloud_name: Optional[str],
    name: Optional[str],
    cloud_id: Optional[str],
    functional_verify: Optional[str],
    strict: bool = False,
) -> bool:
    if cloud_name and name and cloud_name != name:
        raise click.ClickException(
            "The positional argument CLOUD_NAME and the keyword argument --name "
            "were both provided. Please only provide one of these two arguments."
        )

    return CloudController().verify_cloud(
        cloud_name=cloud_name or name,
        cloud_id=cloud_id,
        functional_verify=functional_verify,
        strict=strict,
    )


@cloud_cli.command(name="edit", help="Edit anyscale cloud resource.")
@click.argument("cloud-name", required=False)
@click.option("--name", "-n", help="Edit cloud by name.", type=str)
@click.option(
    "--cloud-id",
    "--id",
    help="Edit cloud by id, alternative to cloud name.",
    required=False,
)
@click.option(
    "--aws-s3-id", help="New S3 bucket ID.", required=False, type=str,
)
@click.option("--aws-efs-id", help="New EFS ID.", required=False, type=str)
@click.option(
    "--aws-efs-mount-target-ip",
    help="New EFS mount target IP.",
    required=False,
    type=str,
)
@click.option(
    "--gcp-filestore-instance-id",
    help="New GCP filestore instance id.",
    required=False,
    type=str,
)
@click.option(
    "--gcp-filestore-location",
    help="New GCP filestore location.",
    required=False,
    type=str,
)
@click.option(
    "--gcp-cloud-storage-bucket-name",
    help="New GCP Cloud storage bucket name.",
    required=False,
    type=str,
)
@click.option(
    "--functional-verify",
    help="Verify the cloud is functional. This will check that the cloud can launch workspace/service.",
    required=False,
    is_flag=False,
    flag_value="workspace",
)
def cloud_edit(  # noqa: PLR0913
    cloud_name: Optional[str],
    name: Optional[str],
    cloud_id: Optional[str],
    aws_s3_id: Optional[str],
    aws_efs_id: Optional[str],
    aws_efs_mount_target_ip: Optional[str],
    gcp_filestore_instance_id: Optional[str],
    gcp_filestore_location: Optional[str],
    gcp_cloud_storage_bucket_name: Optional[str],
    functional_verify: Optional[str],
) -> None:
    if cloud_name and name and cloud_name != name:
        raise click.ClickException(
            "The positional argument CLOUD_NAME and the keyword argument --name "
            "were both provided. Please only provide one of these two arguments."
        )
    if any(
        [
            aws_s3_id,
            aws_efs_id,
            aws_efs_mount_target_ip,
            gcp_filestore_instance_id,
            gcp_filestore_location,
            gcp_cloud_storage_bucket_name,
        ]
    ):
        if any([gcp_filestore_instance_id, gcp_filestore_location]) and not all(
            [gcp_filestore_instance_id, gcp_filestore_location]
        ):
            # Make sure both gcp_filestore_instance_id and gcp_filestore_location are provided if you want to edit filestore.
            raise click.ClickException(
                "Please provide both --gcp-filestore-instance-id and --gcp-filestore-location if you want to edit filestore."
            )
        CloudController().edit_cloud(
            cloud_name=cloud_name or name,
            cloud_id=cloud_id,
            aws_s3_id=aws_s3_id,
            aws_efs_id=aws_efs_id,
            aws_efs_mount_target_ip=aws_efs_mount_target_ip,
            gcp_filestore_instance_id=gcp_filestore_instance_id,
            gcp_filestore_location=gcp_filestore_location,
            gcp_cloud_storage_bucket_name=gcp_cloud_storage_bucket_name,
            functional_verify=functional_verify,
        )
    else:
        raise click.ClickException(
            "Please provide at least one of the following arguments: --aws-s3-id, --aws-efs-id, --aws-efs-mount-target-ip, --gcp-filestore-instance-id, --gcp-filestore-location, --gcp-cloud-storage-bucket-name."
        )



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class DecoratedProductionJobStateTransition(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'id': 'str',
        'state_transitioned_at': 'datetime',
        'current_state': 'HaJobStates',
        'goal_state': 'HaJobGoalStates',
        'error': 'str',
        'operation_message': 'str',
        'cluster_id': 'str',
        'cluster': 'MiniCluster'
    }

    attribute_map = {
        'id': 'id',
        'state_transitioned_at': 'state_transitioned_at',
        'current_state': 'current_state',
        'goal_state': 'goal_state',
        'error': 'error',
        'operation_message': 'operation_message',
        'cluster_id': 'cluster_id',
        'cluster': 'cluster'
    }

    def __init__(self, id=None, state_transitioned_at=None, current_state=None, goal_state=None, error=None, operation_message=None, cluster_id=None, cluster=None, local_vars_configuration=None):  # noqa: E501
        """DecoratedProductionJobStateTransition - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._id = None
        self._state_transitioned_at = None
        self._current_state = None
        self._goal_state = None
        self._error = None
        self._operation_message = None
        self._cluster_id = None
        self._cluster = None
        self.discriminator = None

        self.id = id
        self.state_transitioned_at = state_transitioned_at
        self.current_state = current_state
        if goal_state is not None:
            self.goal_state = goal_state
        if error is not None:
            self.error = error
        if operation_message is not None:
            self.operation_message = operation_message
        if cluster_id is not None:
            self.cluster_id = cluster_id
        if cluster is not None:
            self.cluster = cluster

    @property
    def id(self):
        """Gets the id of this DecoratedProductionJobStateTransition.  # noqa: E501

        The id of this job state transition  # noqa: E501

        :return: The id of this DecoratedProductionJobStateTransition.  # noqa: E501
        :rtype: str
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this DecoratedProductionJobStateTransition.

        The id of this job state transition  # noqa: E501

        :param id: The id of this DecoratedProductionJobStateTransition.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and id is None:  # noqa: E501
            raise ValueError("Invalid value for `id`, must not be `None`")  # noqa: E501

        self._id = id

    @property
    def state_transitioned_at(self):
        """Gets the state_transitioned_at of this DecoratedProductionJobStateTransition.  # noqa: E501

        The last time the state of this job was updated. This includes updates to the state and to the goal state  # noqa: E501

        :return: The state_transitioned_at of this DecoratedProductionJobStateTransition.  # noqa: E501
        :rtype: datetime
        """
        return self._state_transitioned_at

    @state_transitioned_at.setter
    def state_transitioned_at(self, state_transitioned_at):
        """Sets the state_transitioned_at of this DecoratedProductionJobStateTransition.

        The last time the state of this job was updated. This includes updates to the state and to the goal state  # noqa: E501

        :param state_transitioned_at: The state_transitioned_at of this DecoratedProductionJobStateTransition.  # noqa: E501
        :type: datetime
        """
        if self.local_vars_configuration.client_side_validation and state_transitioned_at is None:  # noqa: E501
            raise ValueError("Invalid value for `state_transitioned_at`, must not be `None`")  # noqa: E501

        self._state_transitioned_at = state_transitioned_at

    @property
    def current_state(self):
        """Gets the current_state of this DecoratedProductionJobStateTransition.  # noqa: E501

        The current state of the job  # noqa: E501

        :return: The current_state of this DecoratedProductionJobStateTransition.  # noqa: E501
        :rtype: HaJobStates
        """
        return self._current_state

    @current_state.setter
    def current_state(self, current_state):
        """Sets the current_state of this DecoratedProductionJobStateTransition.

        The current state of the job  # noqa: E501

        :param current_state: The current_state of this DecoratedProductionJobStateTransition.  # noqa: E501
        :type: HaJobStates
        """
        if self.local_vars_configuration.client_side_validation and current_state is None:  # noqa: E501
            raise ValueError("Invalid value for `current_state`, must not be `None`")  # noqa: E501

        self._current_state = current_state

    @property
    def goal_state(self):
        """Gets the goal_state of this DecoratedProductionJobStateTransition.  # noqa: E501

        The goal state of the job  # noqa: E501

        :return: The goal_state of this DecoratedProductionJobStateTransition.  # noqa: E501
        :rtype: HaJobGoalStates
        """
        return self._goal_state

    @goal_state.setter
    def goal_state(self, goal_state):
        """Sets the goal_state of this DecoratedProductionJobStateTransition.

        The goal state of the job  # noqa: E501

        :param goal_state: The goal_state of this DecoratedProductionJobStateTransition.  # noqa: E501
        :type: HaJobGoalStates
        """

        self._goal_state = goal_state

    @property
    def error(self):
        """Gets the error of this DecoratedProductionJobStateTransition.  # noqa: E501

        An error message that occurred in this job state transition  # noqa: E501

        :return: The error of this DecoratedProductionJobStateTransition.  # noqa: E501
        :rtype: str
        """
        return self._error

    @error.setter
    def error(self, error):
        """Sets the error of this DecoratedProductionJobStateTransition.

        An error message that occurred in this job state transition  # noqa: E501

        :param error: The error of this DecoratedProductionJobStateTransition.  # noqa: E501
        :type: str
        """

        self._error = error

    @property
    def operation_message(self):
        """Gets the operation_message of this DecoratedProductionJobStateTransition.  # noqa: E501

        The logging message for this job state transition  # noqa: E501

        :return: The operation_message of this DecoratedProductionJobStateTransition.  # noqa: E501
        :rtype: str
        """
        return self._operation_message

    @operation_message.setter
    def operation_message(self, operation_message):
        """Sets the operation_message of this DecoratedProductionJobStateTransition.

        The logging message for this job state transition  # noqa: E501

        :param operation_message: The operation_message of this DecoratedProductionJobStateTransition.  # noqa: E501
        :type: str
        """

        self._operation_message = operation_message

    @property
    def cluster_id(self):
        """Gets the cluster_id of this DecoratedProductionJobStateTransition.  # noqa: E501

        The id of the cluster the job is running on  # noqa: E501

        :return: The cluster_id of this DecoratedProductionJobStateTransition.  # noqa: E501
        :rtype: str
        """
        return self._cluster_id

    @cluster_id.setter
    def cluster_id(self, cluster_id):
        """Sets the cluster_id of this DecoratedProductionJobStateTransition.

        The id of the cluster the job is running on  # noqa: E501

        :param cluster_id: The cluster_id of this DecoratedProductionJobStateTransition.  # noqa: E501
        :type: str
        """

        self._cluster_id = cluster_id

    @property
    def cluster(self):
        """Gets the cluster of this DecoratedProductionJobStateTransition.  # noqa: E501


        :return: The cluster of this DecoratedProductionJobStateTransition.  # noqa: E501
        :rtype: MiniCluster
        """
        return self._cluster

    @cluster.setter
    def cluster(self, cluster):
        """Sets the cluster of this DecoratedProductionJobStateTransition.


        :param cluster: The cluster of this DecoratedProductionJobStateTransition.  # noqa: E501
        :type: MiniCluster
        """

        self._cluster = cluster

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, DecoratedProductionJobStateTransition):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, DecoratedProductionJobStateTransition):
            return True

        return self.to_dict() != other.to_dict()



import copy
import json
from typing import Any, List
from unittest.mock import Mock

import boto3
from moto import mock_iam, mock_s3
import pytest

from anyscale.aws_iam_policies import AMAZON_S3_FULL_ACCESS_POLICY_ARN
from anyscale.utils.s3 import (
    _verify_identity_based_s3_access,
    _verify_resource_based_s3_access,
    REQUIRED_S3_PERMISSIONS,
    verify_s3_access,
)


"""
Negative Test Cases
* Role with Wrong S3 Bucket permissions
"""

_POLICY_TEMPLATE = {
    "Version": "2012-10-17",
    "Statement": [{"Effect": "Allow", "Action": list(REQUIRED_S3_PERMISSIONS)}],
}

_BUCKET_NAME = "bucket-test"


def _create_role() -> Any:
    _name = "our-role"
    boto3.client("iam").create_role(RoleName=_name, AssumeRolePolicyDocument="{}")
    return boto3.resource("iam").Role(_name)


def _create_bucket() -> Any:
    boto3.client("s3").create_bucket(Bucket=_BUCKET_NAME)
    return boto3.resource("s3").Bucket(_BUCKET_NAME)


def _add_policy(role: Any, resource: str, inline: bool, as_list: bool):
    policy_document = copy.deepcopy(_POLICY_TEMPLATE)
    policy_document["Statement"][0]["Resource"] = [resource] if as_list else resource  # type: ignore
    if inline:
        boto3.client("iam").put_role_policy(
            RoleName=role.name,
            PolicyName="inline-iam-s3-specific",
            PolicyDocument=json.dumps(policy_document),
        )
    else:
        policy = boto3.client("iam").create_policy(
            PolicyName="iam-s3-specific", PolicyDocument=json.dumps(policy_document)
        )["Policy"]["Arn"]
        role.attach_policy(PolicyArn=policy)


@mock_iam
@mock_s3
@pytest.mark.parametrize(
    "resource",
    [
        pytest.param("arn:aws:s3:::" + _BUCKET_NAME, id="S3SpecificPermissions"),
        pytest.param("*", id="S3GenericPermissions"),
    ],
)
@pytest.mark.parametrize("inline", [True, False])
@pytest.mark.parametrize("as_list", [True, False])
def test_correct_policy_on_role(resource: str, inline: bool, as_list: bool):
    role = _create_role()
    bucket = _create_bucket()
    _add_policy(role, resource, inline, as_list)
    mock_logger = Mock()

    assert _verify_identity_based_s3_access(boto3.Session(), bucket, role, mock_logger)
    mock_logger.info.assert_not_called()


@mock_iam
@mock_s3
def test_s3_full_access_on_role():
    role = _create_role()
    role.attach_policy(PolicyArn=AMAZON_S3_FULL_ACCESS_POLICY_ARN)
    bucket = _create_bucket()
    mock_logger = Mock()

    assert _verify_identity_based_s3_access(boto3.Session(), bucket, role, mock_logger)
    mock_logger.info.assert_not_called()


@mock_iam
@mock_s3
@pytest.mark.parametrize(
    "allow_actions",
    [
        pytest.param(["s3:*"], id="S3_Star"),
        pytest.param(list(REQUIRED_S3_PERMISSIONS), id="S3_Specific"),
    ],
)
def test_correct_policy_on_bucket(allow_actions: List[str]):
    role = _create_role()
    bucket = _create_bucket()

    policy_document = copy.deepcopy(_POLICY_TEMPLATE)
    policy_document["Statement"][0]["Resource"] = [  # type: ignore
        f"arn:aws:s3:::{bucket.name}",
        f"arn:aws:s3:::{bucket.name}/*",
    ]
    policy_document["Statement"][0]["Principal"] = {"AWS": role.arn}  # type: ignore
    policy_document["Statement"][0]["Actions"] = allow_actions  # type: ignore
    bucket.Policy().put(Policy=json.dumps(policy_document))

    mock_logger = Mock()
    assert _verify_resource_based_s3_access(bucket, role, mock_logger)
    mock_logger.info.assert_not_called()


@mock_s3
@mock_iam
def test_no_permissions():
    bucket = _create_bucket()
    role = _create_role()
    mock_logger = Mock()
    assert not verify_s3_access(boto3.Session(), bucket, role, mock_logger)
    mock_logger.info.assert_not_called()


@mock_iam
@mock_s3
def test_policy_on_bucket_missing_perms():
    role = _create_role()
    bucket = _create_bucket()

    policy_document = copy.deepcopy(_POLICY_TEMPLATE)
    policy_document["Statement"][0]["Resource"] = [  # type: ignore
        f"arn:aws:s3:::{bucket.name}",
        f"arn:aws:s3:::{bucket.name}/*",
    ]
    policy_document["Statement"][0]["Principal"] = {"AWS": role.arn}  # type: ignore
    policy_document["Statement"][0]["Action"] = list(REQUIRED_S3_PERMISSIONS)[:-1]  # type: ignore
    bucket.Policy().put(Policy=json.dumps(policy_document))

    assert not verify_s3_access(boto3.Session(), bucket, role, Mock())

    mock_logger = Mock()
    assert not _verify_resource_based_s3_access(bucket, role, mock_logger)
    mock_logger.info.assert_called_once()


@mock_iam
@mock_s3
def test_policy_on_role_missing_perms():
    role = _create_role()
    bucket = _create_bucket()

    policy_document = copy.deepcopy(_POLICY_TEMPLATE)
    policy_document["Statement"][0]["Resource"] = [  # type: ignore
        f"arn:aws:s3:::{bucket.name}",
        f"arn:aws:s3:::{bucket.name}/*",
    ]
    policy_document["Statement"][0]["Action"] = list(REQUIRED_S3_PERMISSIONS)[:-1]  # type: ignore
    boto3.client("iam").put_role_policy(
        RoleName=role.name,
        PolicyName="inline-iam-s3-specific",
        PolicyDocument=json.dumps(policy_document),
    )

    assert not verify_s3_access(boto3.Session(), bucket, role, Mock())

    mock_logger = Mock()
    assert not _verify_identity_based_s3_access(
        boto3.Session(), bucket, role, mock_logger
    )
    mock_logger.info.assert_called_once()


@mock_iam
@mock_s3
@pytest.mark.parametrize("inline", [True, False])
@pytest.mark.parametrize("as_list", [True, False])
def test_policy_on_role_for_wrong_bucket(inline: bool, as_list: bool):
    role = _create_role()
    bucket = _create_bucket()
    _add_policy(role, "arn:aws:s3:::incorrect-bucket-name", inline, as_list)

    mock_logger = Mock()
    assert not verify_s3_access(boto3.Session(), bucket, role, mock_logger)
    mock_logger.info.assert_not_called()


@mock_iam
@mock_s3
def test_policy_on_bucket_for_wrong_role():
    role = _create_role()
    bucket = _create_bucket()

    policy_document = copy.deepcopy(_POLICY_TEMPLATE)
    policy_document["Statement"][0]["Resource"] = [  # type: ignore
        f"arn:aws:s3:::{bucket.name}",
        f"arn:aws:s3:::{bucket.name}/*",
    ]
    policy_document["Statement"][0]["Principal"] = {"AWS": "something_random"}  # type: ignore
    bucket.Policy().put(Policy=json.dumps(policy_document))

    mock_logger = Mock()
    assert not verify_s3_access(boto3.Session(), bucket, role, mock_logger)
    mock_logger.info.assert_not_called()



import os
from typing import Any, Dict
import unittest
from unittest.mock import Mock

import pytest

import anyscale.job


ha_job_id = "ha_job_id"
ray_job_id = "ray_job_id"
should_succeed = "should_succeed"
false_str = "False"
true_str = "True"


@pytest.mark.parametrize(
    "env",
    [
        {
            anyscale.job.ANYSCALE_HA_JOB_ID: "",
            anyscale.job.ANYSCALE_RAY_JOB_SUBMISSION_ID: "",
            should_succeed: false_str,  # this is necessary because the value type must be a string
        },
        {
            anyscale.job.ANYSCALE_HA_JOB_ID: ha_job_id,
            anyscale.job.ANYSCALE_RAY_JOB_SUBMISSION_ID: "",
            should_succeed: false_str,
        },
        {
            anyscale.job.ANYSCALE_HA_JOB_ID: "",
            anyscale.job.ANYSCALE_RAY_JOB_SUBMISSION_ID: ray_job_id,
            should_succeed: false_str,
        },
        {
            anyscale.job.ANYSCALE_HA_JOB_ID: ha_job_id,
            anyscale.job.ANYSCALE_RAY_JOB_SUBMISSION_ID: ray_job_id,
            should_succeed: true_str,
        },
    ],
)
def test_get_env_vars(env):
    with unittest.mock.patch.dict(os.environ, env, clear=True):
        ids = anyscale.job._get_ha_job_id_from_environment()
        assert (
            (ha_job_id, ray_job_id) if env[should_succeed] == true_str else ids is None
        )


def test_submit_output_not_dict():
    anyscale.job._submit_raw_output = Mock()
    anyscale.job.logger = Mock()

    # Don't submit the output unless it is a dict
    anyscale.job.output("abcd", _fail_on_error=True)  # type: ignore
    anyscale.job.output([], _fail_on_error=True)  # type: ignore
    anyscale.job._submit_raw_output.assert_not_called()


def test_submit_output_not_serializable():
    anyscale.job._submit_raw_output = Mock()
    anyscale.job.logger = Mock()
    # Don't fail if the output is not JSON serializable
    circular: Dict[str, Any] = {}
    circular["key"] = circular
    anyscale.job.output(circular, _fail_on_error=True)
    anyscale.job._submit_raw_output.assert_not_called()
    anyscale.job.logger.warn.assert_called_once_with(
        "Failed to serialize the job output to JSON. Output will not be recorded."
    )


def test_submit_output_env_not_set():
    anyscale.job._submit_raw_output = Mock()
    anyscale.job.logger = Mock()
    # Don't submit if the env isn't set
    data = {"key": "value"}
    os.environ[anyscale.job.ANYSCALE_HA_JOB_ID] = ""
    os.environ[anyscale.job.ANYSCALE_RAY_JOB_SUBMISSION_ID] = ""
    anyscale.job.output(data, _fail_on_error=True)
    anyscale.job._submit_raw_output.assert_not_called()


def test_submit_output_api_exception():
    anyscale.job._submit_raw_output = Mock()
    anyscale.job.logger = Mock()

    # Do not raise an exception if the API returns an exception
    anyscale.job._submit_raw_output.side_effect = lambda *args, **kwargs: Exception(
        "test"
    )
    data = {"key": "value"}
    os.environ[anyscale.job.ANYSCALE_HA_JOB_ID] = ha_job_id
    os.environ[anyscale.job.ANYSCALE_RAY_JOB_SUBMISSION_ID] = ray_job_id
    anyscale.job.output(data, _fail_on_error=True)
    anyscale.job._submit_raw_output.assert_called_once_with(
        ha_job_id, ray_job_id, data, api_client=None
    )


def test_submit_output():
    anyscale.job._submit_raw_output = Mock()
    anyscale.job.logger = Mock()
    anyscale.job._submit_raw_output.reset_mock()
    # Do submit
    data = {"key": "value"}
    os.environ[anyscale.job.ANYSCALE_HA_JOB_ID] = ha_job_id
    os.environ[anyscale.job.ANYSCALE_RAY_JOB_SUBMISSION_ID] = ray_job_id
    anyscale.job.output(data, _fail_on_error=True)
    anyscale.job._submit_raw_output.assert_called_once_with(
        ha_job_id, ray_job_id, data, api_client=None
    )



import json
from typing import Dict, List, Optional
from unittest.mock import Mock, mock_open, patch

import click
import pytest

from anyscale.controllers.cluster_controller import ClusterController
from anyscale.sdk.anyscale_client import StartClusterOptions
from anyscale.sdk.anyscale_client.models.cluster_compute_config import (
    ClusterComputeConfig,
)


@pytest.mark.parametrize(
    "cluster",
    [
        Mock(state="Running", cluster_environment_build_id="build1"),
        Mock(state="Terminated", cluster_environment_build_id="build1"),
    ],
)
@pytest.mark.parametrize(
    "build", [Mock(id="build1", revision=1), Mock(id="build2", revision=2)]
)
@pytest.mark.parametrize("cluster_compute", [Mock(id="compute1"), Mock(id="compute2")])
@pytest.mark.parametrize("passed_cluster_env", [True, False])
@pytest.mark.parametrize("passed_cluster_compute", [True, False])
def test_check_needs_start(
    cluster: Mock,
    build: Mock,
    cluster_compute: Mock,
    passed_cluster_env: bool,
    passed_cluster_compute: bool,
    mock_auth_api_client,
) -> None:
    cluster_controller = ClusterController()

    needs_start = cluster_controller._check_needs_start(
        cluster, build, cluster_compute, passed_cluster_env, passed_cluster_compute
    )
    if cluster.state != "Running":
        assert needs_start
    elif not passed_cluster_env and not passed_cluster_compute:
        assert not needs_start
    elif passed_cluster_env and cluster.cluster_environment_build_id != build.id:
        assert needs_start
    elif passed_cluster_compute and cluster.cluster_compute_id != cluster_compute.id:
        assert needs_start


def test_get_project_id_and_cluster_name_given_cluster_id(mock_auth_api_client):
    cluster_controller = ClusterController()

    # Test passing in cluster id only
    mock_cluster = Mock(id="mock_cluster_id", project_id="mock_project_id")
    mock_cluster.name = "mock_cluster_name"
    cluster_controller.anyscale_api_client.get_cluster = Mock(
        return_value=Mock(result=mock_cluster)
    )
    assert cluster_controller._get_project_id_and_cluster_name(
        cluster_id="mock_cluster_id",
        project_id=None,
        cluster_name=None,
        project_name=None,
        cloud_id=None,
    ) == ("mock_project_id", "mock_cluster_name")

    # Test passing in project id
    cluster_controller._get_or_generate_cluster_name = Mock(  # type: ignore
        return_value="mock_cluster_name1"
    )
    mock_get_and_validate_project_id = Mock(return_value="mock_project_id1")
    with patch.multiple(
        "anyscale.controllers.cluster_controller",
        get_and_validate_project_id=mock_get_and_validate_project_id,
    ):
        assert cluster_controller._get_project_id_and_cluster_name(
            cluster_id=None,
            project_id="mock_project_id1",
            cluster_name=None,
            project_name=None,
            cloud_id=None,
        ) == ("mock_project_id1", "mock_cluster_name1")
        mock_get_and_validate_project_id.assert_called_once_with(
            project_id="mock_project_id1",
            project_name=None,
            parent_cloud_id=None,
            api_client=cluster_controller.api_client,
            anyscale_api_client=cluster_controller.anyscale_api_client,
        )

    # Test passing in project name
    cluster_controller._get_or_generate_cluster_name = Mock(  # type: ignore
        return_value="mock_cluster_name1"
    )
    mock_get_and_validate_project_id.reset_mock()
    with patch.multiple(
        "anyscale.controllers.cluster_controller",
        get_and_validate_project_id=mock_get_and_validate_project_id,
    ):
        assert cluster_controller._get_project_id_and_cluster_name(
            cluster_id=None,
            project_id=None,
            cluster_name=None,
            project_name="mock_project_name1",
            cloud_id=None,
        ) == ("mock_project_id1", "mock_cluster_name1")
        mock_get_and_validate_project_id.assert_called_once_with(
            project_id=None,
            project_name="mock_project_name1",
            parent_cloud_id=None,
            api_client=cluster_controller.api_client,
            anyscale_api_client=cluster_controller.anyscale_api_client,
        )

    # Test passing in cloud_id
    mock_get_and_validate_project_id.reset_mock()
    with patch.multiple(
        "anyscale.controllers.cluster_controller",
        get_and_validate_project_id=mock_get_and_validate_project_id,
    ):
        assert cluster_controller._get_project_id_and_cluster_name(
            cluster_id=None,
            project_id=None,
            cluster_name=None,
            project_name=None,
            cloud_id="mock_cloud_id",
        ) == ("mock_project_id1", "mock_cluster_name1")
        mock_get_and_validate_project_id.assert_called_once_with(
            project_id=None,
            project_name=None,
            parent_cloud_id="mock_cloud_id",
            api_client=cluster_controller.api_client,
            anyscale_api_client=cluster_controller.anyscale_api_client,
        )


def test_get_cluster_env_and_build(mock_auth_api_client):
    cluster_controller = ClusterController()
    mock_get_build_from_cluster_env_identifier = Mock(return_value=Mock(id="build_id1"))
    mock_get_default_cluster_env_build = Mock(return_value=Mock(id="default_build_id"))

    with patch.multiple(
        "anyscale.controllers.cluster_controller",
        get_build_from_cluster_env_identifier=mock_get_build_from_cluster_env_identifier,
        get_default_cluster_env_build=mock_get_default_cluster_env_build,
    ):
        assert (
            cluster_controller._get_cluster_env_and_build(None)[1].id
            == "default_build_id"
        )
        assert (
            cluster_controller._get_cluster_env_and_build("mock_cluster_env")[1].id
            == "build_id1"
        )


def test_load_cluster_compute(mock_auth_api_client):
    cluster_controller = ClusterController()
    cluster_compute_dict = {
        "cloud_id": "mock_cloud_id",
        "max_workers": 10,
        "region": "us-west-2",
        "allowed_azs": ["us-west-2a", "us-west-2b"],
        "head_node_type": {
            "name": "head-1",
            "instance_type": "m5.2xlarge",
            "aws_advanced_configurations": None,
            "gcp_advanced_configurations": None,
            "resources": None,
        },
        "worker_node_types": [],
    }
    with patch(
        "builtins.open", mock_open(read_data=json.dumps(cluster_compute_dict)),
    ):
        assert cluster_controller._load_cluster_compute(
            "cluster_compute_file",
        ) == ClusterComputeConfig(**cluster_compute_dict)


def test_get_cluster_compute(mock_auth_api_client):
    cluster_controller = ClusterController()
    mock_get_cluster_compute_from_name = Mock(
        return_value=Mock(id="new_cluster_compute_1")
    )
    cluster_controller.anyscale_api_client.create_cluster_compute = Mock(
        return_value=Mock(result=Mock(id="create_cluster_compute_2"))
    )
    mock_get_default_cluster_compute = Mock(
        return_value=Mock(id="default_cluster_compute_3_and_5")
    )
    mock_existing_cluster = Mock(cluster_compute_id="mock_existing_cluster_compute_id")
    cluster_controller.anyscale_api_client.get_cluster_compute = Mock(
        return_value=Mock(result=Mock(id="existing_cluster_compute_4"))
    )

    with patch(
        "builtins.open",
        mock_open(
            read_data=json.dumps(
                {
                    "cloud_id": "mock_cloud_id",
                    "region": "mock_region",
                    "head_node_type": "mock_head_node_type",
                    "worker_node_types": "mock_worker_node_types",
                }
            )
        ),
    ), patch.multiple(
        "anyscale.controllers.cluster_controller",
        get_cluster_compute_from_name=mock_get_cluster_compute_from_name,
        get_default_cluster_compute=mock_get_default_cluster_compute,
    ):
        assert (
            cluster_controller._get_cluster_compute(
                "cluster_compute_name1", None, None, None, "mock_project_id"
            ).id
            == "new_cluster_compute_1"
        )
        assert (
            cluster_controller._get_cluster_compute(
                None, "cluster_compute_file2", None, None, "mock_project_id"
            ).id
            == "create_cluster_compute_2"
        )
        assert (
            cluster_controller._get_cluster_compute(
                None, None, "cloud_name3", None, "mock_project_id"
            ).id
            == "default_cluster_compute_3_and_5"
        )
        assert (
            cluster_controller._get_cluster_compute(
                None, None, None, mock_existing_cluster, "mock_project_id"
            ).id
            == "existing_cluster_compute_4"
        )
        assert (
            cluster_controller._get_cluster_compute(
                None, None, None, None, "mock_project_id"
            ).id
            == "default_cluster_compute_3_and_5"
        )


@pytest.mark.parametrize("needs_start", [True, False])
@pytest.mark.parametrize("cluster_exists", [True, False])
def test_start(mock_auth_api_client, needs_start: bool, cluster_exists: bool) -> None:
    cluster_controller = ClusterController()
    mock_build = Mock(id="build_id")
    mock_cluster_compute = Mock(id="cluster_compute_id")
    cluster_controller._get_project_id_and_cluster_name = Mock(  # type: ignore
        return_value=("mock_project_id", "mock_cluster_name")
    )
    mock_existing_cluster = Mock() if cluster_exists else None
    cluster_list = [mock_existing_cluster] if cluster_exists else []
    cluster_controller.anyscale_api_client.search_clusters = Mock(
        return_value=Mock(results=cluster_list)
    )
    cluster_controller._get_cluster_env_and_build = Mock(  # type: ignore
        return_value=(Mock(id="cluster_env_id"), mock_build)
    )
    cluster_controller._get_cluster_compute = Mock(return_value=mock_cluster_compute)  # type: ignore
    cluster_controller._create_or_update_cluster_data = Mock(  # type: ignore
        return_value=(Mock(id="cluster_id"), needs_start)
    )
    mock_user_service_access = "private"
    mock_cloud = Mock(id="mock_cloud_id")
    mock_cloud.name = "mock_cloud_name"
    cluster_controller.anyscale_api_client.get_cloud = Mock(
        return_value=Mock(result=mock_cloud)
    )
    mock_get_selected_cloud_id_or_default = Mock(return_value="mock_cloud_id")
    mock_get_cluster_compute_from_name = Mock(
        return_value=Mock(id="mock_cluster_compute_id")
    )
    with patch.multiple(
        "anyscale.controllers.cluster_controller",
        wait_for_session_start=Mock(),
        get_selected_cloud_id_or_default=mock_get_selected_cloud_id_or_default,
        get_cluster_compute_from_name=mock_get_cluster_compute_from_name,
    ):
        cluster_controller.start(
            cluster_name=None,
            cluster_id="cluster_id",
            cluster_env_name="cluster_env_name",
            docker=None,
            python_version=None,
            ray_version=None,
            cluster_compute_name="cluster_compute_name",
            cluster_compute_file=None,
            cloud_name=None,
            idle_timeout=None,
            project_id=None,
            project_name=None,
            user_service_access=mock_user_service_access,
        )

    cluster_controller._get_project_id_and_cluster_name.assert_called_once_with(
        "cluster_id", None, None, None, cloud_id="mock_cloud_id",
    )
    mock_get_selected_cloud_id_or_default.assert_called_once_with(
        cluster_controller.api_client,
        cluster_controller.anyscale_api_client,
        cluster_compute_id="mock_cluster_compute_id",
        cluster_compute_config=None,
        cloud_id=None,
        cloud_name=None,
    )
    mock_get_cluster_compute_from_name.assert_called_once_with(
        "cluster_compute_name", cluster_controller.api_client
    )
    cluster_controller.anyscale_api_client.search_clusters.assert_called_once_with(
        {
            "project_id": "mock_project_id",
            "name": {"equals": "mock_cluster_name"},
            "archive_status": "ALL",
        }
    )
    cluster_controller._get_cluster_env_and_build.assert_called_once_with(
        "cluster_env_name"
    )
    cluster_controller._get_cluster_compute.assert_called_once_with(
        "cluster_compute_name", None, None, mock_existing_cluster, "mock_project_id"
    )
    cluster_controller._create_or_update_cluster_data.assert_called_once_with(
        mock_existing_cluster,
        "mock_cluster_name",
        "mock_project_id",
        mock_build,
        mock_cluster_compute,
        True,
        True,
        None,
        mock_user_service_access,
    )

    if needs_start:
        cluster_controller.anyscale_api_client.start_cluster.assert_called_once_with(
            "cluster_id",
            StartClusterOptions(
                cluster_environment_build_id=mock_build.id,
                cluster_compute_id=mock_cluster_compute.id,
            ),
        )


@pytest.mark.parametrize("cluster_list", [[], [Mock(id="mock_cluster_id")]])
@pytest.mark.parametrize(
    "terminate_args",
    [
        {
            "cluster_name": "cluster_name",
            "cluster_id": None,
            "project_id": "project_id",
            "project_name": None,
            "cloud_name": None,
            "cloud_id": None,
        },
        {
            "cluster_name": "cluster_name",
            "cluster_id": None,
            "project_id": None,
            "project_name": None,
            "cloud_name": "cloud_name",
            "cloud_id": None,
        },
    ],
)
def test_terminate(
    mock_auth_api_client,
    cluster_list: List[Mock],
    terminate_args: Dict[str, Optional[str]],
) -> None:
    cluster_controller = ClusterController()
    cluster_controller._get_project_id_and_cluster_name = Mock(  # type: ignore
        return_value=("mock_project_id", "mock_cluster_name")
    )
    cluster_controller.anyscale_api_client.search_clusters = Mock(  # type: ignore
        return_value=Mock(results=cluster_list)
    )
    mock_found_cloud_id = "mock_cloud_id" if terminate_args["cloud_name"] else None
    cluster_controller.api_client.find_cloud_by_name_api_v2_clouds_find_by_name_post = Mock(
        return_value=Mock(result=Mock(id=mock_found_cloud_id))
    )

    with pytest.raises(click.ClickException):
        cluster_controller.terminate(None, None, None, None, None, None)
    with pytest.raises(click.ClickException):
        cluster_controller.terminate(
            "cluster_name", "cluster_id", None, None, None, None
        )
    with pytest.raises(click.ClickException):
        cluster_controller.terminate(
            "cluster_name", None, None, None, "cloud_name", "cloud_id"
        )
    with pytest.raises(click.ClickException):
        cluster_controller.api_client.check_is_feature_flag_on_api_v2_userinfo_check_is_feature_flag_on_get = Mock(
            return_value=Mock(result=Mock(is_on=True))
        )
        cluster_controller.terminate("cluster_name", None, None, None, None, None)
    if len(cluster_list) == 0:
        with pytest.raises(click.ClickException):
            cluster_controller.terminate(**terminate_args)
    else:
        cluster_controller._validate_args_to_find_cluster = Mock()  # type: ignore
        cluster_controller.terminate(**terminate_args)
        cluster_controller._get_project_id_and_cluster_name.assert_called_once_with(
            terminate_args["cluster_id"],
            terminate_args["project_id"],
            terminate_args["cluster_name"],
            terminate_args["project_name"],
            cloud_id=mock_found_cloud_id,
        )
        cluster_controller._validate_args_to_find_cluster.assert_called_once_with(
            terminate_args["cluster_name"],
            terminate_args["cluster_id"],
            terminate_args["project_name"],
            terminate_args["project_id"],
            terminate_args["cloud_name"],
            terminate_args["cloud_id"],
        )
        cluster_controller.anyscale_api_client.search_clusters.assert_called_once_with(
            {"project_id": "mock_project_id", "name": {"equals": "mock_cluster_name"}}
        )
        cluster_controller.anyscale_api_client.terminate_cluster.assert_called_once_with(
            "mock_cluster_id", {}
        )


@pytest.mark.parametrize("cluster_list", [[], [Mock(id="mock_cluster_id")]])
@pytest.mark.parametrize(
    "archive_args",
    [
        {
            "cluster_name": "cluster_name",
            "cluster_id": None,
            "project_id": "project_id",
            "project_name": None,
            "cloud_name": None,
            "cloud_id": None,
        },
        {
            "cluster_name": "cluster_name",
            "cluster_id": None,
            "project_id": None,
            "project_name": None,
            "cloud_name": "cloud_name",
            "cloud_id": None,
        },
    ],
)
def test_archive(
    mock_auth_api_client,
    cluster_list: List[Mock],
    archive_args: Dict[str, Optional[str]],
) -> None:
    cluster_controller = ClusterController()
    cluster_controller._get_project_id_and_cluster_name = Mock(  # type: ignore
        return_value=("mock_project_id", "mock_cluster_name")
    )
    cluster_controller.anyscale_api_client.search_clusters = Mock(  # type: ignore
        return_value=Mock(results=cluster_list)
    )
    mock_found_cloud_id = "mock_cloud_id" if archive_args["cloud_name"] else None
    cluster_controller.api_client.find_cloud_by_name_api_v2_clouds_find_by_name_post = Mock(
        return_value=Mock(result=Mock(id=mock_found_cloud_id))
    )
    with pytest.raises(click.ClickException):
        cluster_controller.archive(
            "cluster_name", None, None, None, "cloud_name", "cloud_id"
        )
    with pytest.raises(click.ClickException):
        cluster_controller.api_client.check_is_feature_flag_on_api_v2_userinfo_check_is_feature_flag_on_get = Mock(
            return_value=Mock(result=Mock(is_on=True))
        )
        cluster_controller.archive("cluster_name", None, None, None, None, None)
    with pytest.raises(click.ClickException):
        cluster_controller.archive(None, None, None, None, None, None)
    with pytest.raises(click.ClickException):
        cluster_controller.archive("cluster_name", "cluster_id", None, None, None, None)

    if len(cluster_list) == 0:
        with pytest.raises(click.ClickException):
            cluster_controller.archive(**archive_args)
    else:
        cluster_controller._validate_args_to_find_cluster = Mock()  # type: ignore
        cluster_controller.archive(**archive_args)
        cluster_controller._get_project_id_and_cluster_name.assert_called_once_with(
            archive_args["cluster_id"],
            archive_args["project_id"],
            archive_args["cluster_name"],
            archive_args["project_name"],
            mock_found_cloud_id,
        )
        cluster_controller._validate_args_to_find_cluster.assert_called_once_with(
            archive_args["cluster_name"],
            archive_args["cluster_id"],
            archive_args["project_name"],
            archive_args["project_id"],
            archive_args["cloud_name"],
            archive_args["cloud_id"],
        )
        cluster_controller.anyscale_api_client.search_clusters.assert_called_once_with(
            {"project_id": "mock_project_id", "name": {"equals": "mock_cluster_name"}}
        )
        cluster_controller.anyscale_api_client.archive_cluster.assert_called_once_with(
            "mock_cluster_id"
        )


@pytest.mark.parametrize("cluster_name", ["mock_cluster_name", None])
@pytest.mark.parametrize("cluster_id", ["mock_cluster_id", None])
@pytest.mark.parametrize("project_name", ["mock_project_name", None])
@pytest.mark.parametrize("project_id", ["mock_project_id", None])
@pytest.mark.parametrize("cloud_name", ["mock_cloud_name", None])
@pytest.mark.parametrize("cloud_id", ["mock_cloud_id", None])
@pytest.mark.parametrize("cloud_isolation_ff_on", [True, False])
def test_validate_args_to_find_cluster(  # noqa: PLR0913
    mock_auth_api_client,
    cluster_name: Optional[str],
    cluster_id: Optional[str],
    project_name: Optional[str],
    project_id: Optional[str],
    cloud_name: Optional[str],
    cloud_id: Optional[str],
    cloud_isolation_ff_on: bool,
):
    cluster_controller = ClusterController()
    cluster_controller.api_client.check_is_feature_flag_on_api_v2_userinfo_check_is_feature_flag_on_get = Mock(
        return_value=Mock(result=Mock(is_on=cloud_isolation_ff_on))
    )
    if (bool(cluster_name) + bool(cluster_id) != 1) or (
        bool(cloud_name) + bool(cloud_id) > 1
    ):
        with pytest.raises(click.ClickException):
            cluster_controller._validate_args_to_find_cluster(
                cluster_name, cluster_id, project_name, project_id, cloud_name, cloud_id
            )
        return
    if (
        cloud_isolation_ff_on
        and cluster_name
        and not (project_name or project_id or cloud_name or cloud_id)
    ):
        with pytest.raises(click.ClickException):
            cluster_controller._validate_args_to_find_cluster(
                cluster_name, cluster_id, project_name, project_id, cloud_name, cloud_id
            )
        return
    # Verify runs successfully
    cluster_controller._validate_args_to_find_cluster(
        cluster_name, cluster_id, project_name, project_id, cloud_name, cloud_id
    )



from unittest.mock import Mock

import pytest

from anyscale.cli_logger import CloudSetupLogger
from anyscale.client.openapi_client.models import (
    CloudAnalyticsEvent,
    CloudAnalyticsEventCloudProviderError,
    CloudAnalyticsEventError,
    CreateAnalyticsEvent,
)
from anyscale.utils.cloud_utils import CloudEventProducer


@pytest.mark.parametrize(
    ("internal_error", "cloud_provider_error"),
    [(True, False), (False, True), (False, False), (True, True)],
)
@pytest.mark.parametrize("exception", [True, False])
def test_cloud_event_producer_produce(internal_error, cloud_provider_error, exception):
    mock_api_client = Mock()
    mock_api_client.produce_analytics_event_api_v2_analytics_post = Mock()
    if exception:
        mock_api_client.produce_analytics_event_api_v2_analytics_post.side_effect = (
            Exception()
        )
    logger = CloudSetupLogger()
    mock_cloud_resource = Mock()
    if cloud_provider_error:
        logger.log_resource_error(mock_cloud_resource, "not_found", 404)
    mock_internal_error = "mock_internal_error" if internal_error else None
    expected_cloud_provider_error = (
        [
            CloudAnalyticsEventCloudProviderError(
                cloud_resource=mock_cloud_resource, error_code="not_found,404",
            )
        ]
        if cloud_provider_error
        else None
    )

    cloud_event_producer = CloudEventProducer(mock_api_client)
    cloud_event_producer.init_trace_context(Mock(), Mock())
    mock_event_name = Mock()
    succeeded = not (internal_error or cloud_provider_error)
    expected_error = None
    if internal_error or cloud_provider_error:
        expected_error = CloudAnalyticsEventError(
            internal_error=mock_internal_error,
            cloud_provider_error=expected_cloud_provider_error,
        )

    # shouldn't throw exceptions
    cloud_event_producer.produce(
        event_name=mock_event_name,
        succeeded=succeeded,
        logger=logger,
        internal_error=mock_internal_error,
    )

    mock_api_client.produce_analytics_event_api_v2_analytics_post.assert_called_once_with(
        CreateAnalyticsEvent(
            cloud_analytics_event=CloudAnalyticsEvent(
                trace_id=cloud_event_producer.trace_id,
                cloud_id=cloud_event_producer.cloud_id,
                succeeded=succeeded,
                command_name=cloud_event_producer.command_name,
                raw_command_input=cloud_event_producer.raw_command_input,
                event_name=mock_event_name,
                error=expected_error,
            )
        )
    )



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class Resources(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'cpu': 'int',
        'gpu': 'int',
        'memory': 'int',
        'object_store_memory': 'int',
        'custom_resources': 'dict(str, int)'
    }

    attribute_map = {
        'cpu': 'cpu',
        'gpu': 'gpu',
        'memory': 'memory',
        'object_store_memory': 'object_store_memory',
        'custom_resources': 'custom_resources'
    }

    def __init__(self, cpu=None, gpu=None, memory=None, object_store_memory=None, custom_resources=None, local_vars_configuration=None):  # noqa: E501
        """Resources - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._cpu = None
        self._gpu = None
        self._memory = None
        self._object_store_memory = None
        self._custom_resources = None
        self.discriminator = None

        if cpu is not None:
            self.cpu = cpu
        if gpu is not None:
            self.gpu = gpu
        if memory is not None:
            self.memory = memory
        if object_store_memory is not None:
            self.object_store_memory = object_store_memory
        if custom_resources is not None:
            self.custom_resources = custom_resources

    @property
    def cpu(self):
        """Gets the cpu of this Resources.  # noqa: E501

        Number of CPUs in this node type. If left blank, Ray may automatically detect it for you; see https://docs.ray.io/en/master/cluster/vms/references/ray-cluster-configuration.html#resources for more.  # noqa: E501

        :return: The cpu of this Resources.  # noqa: E501
        :rtype: int
        """
        return self._cpu

    @cpu.setter
    def cpu(self, cpu):
        """Sets the cpu of this Resources.

        Number of CPUs in this node type. If left blank, Ray may automatically detect it for you; see https://docs.ray.io/en/master/cluster/vms/references/ray-cluster-configuration.html#resources for more.  # noqa: E501

        :param cpu: The cpu of this Resources.  # noqa: E501
        :type: int
        """

        self._cpu = cpu

    @property
    def gpu(self):
        """Gets the gpu of this Resources.  # noqa: E501

        Number of GPUs in this node type. If left blank, Ray may automatically detect it for you; see https://docs.ray.io/en/master/cluster/vms/references/ray-cluster-configuration.html#resources for more.  # noqa: E501

        :return: The gpu of this Resources.  # noqa: E501
        :rtype: int
        """
        return self._gpu

    @gpu.setter
    def gpu(self, gpu):
        """Sets the gpu of this Resources.

        Number of GPUs in this node type. If left blank, Ray may automatically detect it for you; see https://docs.ray.io/en/master/cluster/vms/references/ray-cluster-configuration.html#resources for more.  # noqa: E501

        :param gpu: The gpu of this Resources.  # noqa: E501
        :type: int
        """

        self._gpu = gpu

    @property
    def memory(self):
        """Gets the memory of this Resources.  # noqa: E501

        Amount of memory to allocate to the Python worker. If left blank, Ray will choose an appropriate amount based on available resources; see https://docs.ray.io/en/master/cluster/vms/references/ray-cluster-configuration.html#resources for more.  # noqa: E501

        :return: The memory of this Resources.  # noqa: E501
        :rtype: int
        """
        return self._memory

    @memory.setter
    def memory(self, memory):
        """Sets the memory of this Resources.

        Amount of memory to allocate to the Python worker. If left blank, Ray will choose an appropriate amount based on available resources; see https://docs.ray.io/en/master/cluster/vms/references/ray-cluster-configuration.html#resources for more.  # noqa: E501

        :param memory: The memory of this Resources.  # noqa: E501
        :type: int
        """

        self._memory = memory

    @property
    def object_store_memory(self):
        """Gets the object_store_memory of this Resources.  # noqa: E501

        The amount of memory in bytes allocated for the Ray object store on this node. If left blank, Ray will choose an appropriate amount; see https://docs.ray.io/en/master/cluster/vms/references/ray-cluster-configuration.html#resources for more.  # noqa: E501

        :return: The object_store_memory of this Resources.  # noqa: E501
        :rtype: int
        """
        return self._object_store_memory

    @object_store_memory.setter
    def object_store_memory(self, object_store_memory):
        """Sets the object_store_memory of this Resources.

        The amount of memory in bytes allocated for the Ray object store on this node. If left blank, Ray will choose an appropriate amount; see https://docs.ray.io/en/master/cluster/vms/references/ray-cluster-configuration.html#resources for more.  # noqa: E501

        :param object_store_memory: The object_store_memory of this Resources.  # noqa: E501
        :type: int
        """

        self._object_store_memory = object_store_memory

    @property
    def custom_resources(self):
        """Gets the custom_resources of this Resources.  # noqa: E501

        Declare custom resources for this node to be used by Ray autoscaler.  # noqa: E501

        :return: The custom_resources of this Resources.  # noqa: E501
        :rtype: dict(str, int)
        """
        return self._custom_resources

    @custom_resources.setter
    def custom_resources(self, custom_resources):
        """Sets the custom_resources of this Resources.

        Declare custom resources for this node to be used by Ray autoscaler.  # noqa: E501

        :param custom_resources: The custom_resources of this Resources.  # noqa: E501
        :type: dict(str, int)
        """

        self._custom_resources = custom_resources

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, Resources):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, Resources):
            return True

        return self.to_dict() != other.to_dict()



import re
from typing import Any, Tuple
from unittest.mock import Mock, patch

from click import ClickException
from google.auth.credentials import AnonymousCredentials
from google.auth.exceptions import DefaultCredentialsError, RefreshError
import pytest

from anyscale.cli_logger import CloudSetupLogger
from anyscale.client.openapi_client.models.gcp_file_store_config import (
    GCPFileStoreConfig,
)
from anyscale.utils.gcp_utils import (
    get_application_default_credentials,
    get_filestore_location_and_instance_id,
    get_gcp_filestore_config,
    get_google_cloud_client_factory,
)


@pytest.mark.parametrize(
    ("filestore_id", "response", "not_found", "expected_result"),
    [
        pytest.param(
            "regional-filestore",
            ("200 OK", "regional_filestore.json"),
            False,
            GCPFileStoreConfig(
                instance_name="projects/anyscale-bridge-deadbeef/locations/us-central1/instances/regional-filestore",
                root_dir="something",
                mount_target_ip="172.22.236.2",
            ),
            id="regional",
        ),
        pytest.param(
            "zonal-filestore",
            ("200 OK", "zonal_filestore.json"),
            False,
            GCPFileStoreConfig(
                instance_name="projects/anyscale-bridge-deadbeef/locations/us-central1/instances/zonal-filestore",
                root_dir="test",
                mount_target_ip="172.27.155.250",
            ),
            id="zonal",
        ),
        pytest.param(
            "regional-filestore",
            ("200 OK", "regional_filestore_wrong_vpc.json"),
            False,
            None,
            id="wrong_vpc",
        ),
        pytest.param(
            "regional-filestore",
            ("200 OK", "regional_filestore_vpc_private_access.json"),
            False,
            GCPFileStoreConfig(
                instance_name="projects/anyscale-bridge-deadbeef/locations/us-central1/instances/regional-filestore",
                root_dir="something",
                mount_target_ip="172.22.236.2",
            ),
            id="regional-private-access",
        ),
        pytest.param(
            "regional-filestore", ("404 Not Found", None), True, None, id="not_found"
        ),
    ],
)
def test_get_gcp_filestore_config(
    setup_mock_server, filestore_id, capsys, response, not_found, expected_result,
):
    factory, tracker = setup_mock_server
    mock_project_id = "anyscale-bridge-deadbeef"
    mock_vpc_name = "vpc_one"
    mock_filestore_region = "us-central1"
    tracker.reset({".*": [response]})
    if expected_result:
        assert (
            get_gcp_filestore_config(
                factory,
                mock_project_id,
                mock_vpc_name,
                mock_filestore_region,
                filestore_id,
                CloudSetupLogger(),
            )
            == expected_result
        )
    else:
        if not_found:
            with pytest.raises(ClickException) as e:
                get_gcp_filestore_config(
                    factory,
                    mock_project_id,
                    mock_vpc_name,
                    mock_filestore_region,
                    filestore_id,
                    CloudSetupLogger(),
                )
            e.match("Could not find Filestore with id")
        else:
            with pytest.raises(ClickException):
                get_gcp_filestore_config(
                    factory,
                    mock_project_id,
                    mock_vpc_name,
                    mock_filestore_region,
                    filestore_id,
                    CloudSetupLogger(),
                )
            _, err = capsys.readouterr()
            assert "This cannot be edited on an existing Filestore instance." in err


def test_get_application_default_credentials_success(capsys):
    mock_run = Mock()
    with patch(
        "anyscale.utils.gcp_utils.google.auth",
        default=Mock(return_value=(Mock(), "project_abc")),
    ), patch("anyscale.utils.gcp_utils.subprocess", run=mock_run):
        credentials, project = get_application_default_credentials(CloudSetupLogger())
        assert credentials is not None
        assert project == "project_abc"

    _, err = capsys.readouterr()
    assert err == ""

    mock_run.assert_not_called()


@pytest.mark.parametrize(
    ("refresh_error", "expected_output"),
    [
        pytest.param(True, "Reauthentication is needed", id="ReAuthNeeded"),
        pytest.param(False, "GCloud", id="NoAuthFound"),
    ],
)
def test_get_application_default_credentials_reauth(
    capsys, refresh_error: bool, expected_output: str
):
    num_calls = 0

    def default(*args: Any, **kwargs: Any) -> Tuple[Any, str]:
        nonlocal num_calls
        num_calls += 1
        if num_calls > 1:
            return Mock(), "project_abc"
        if refresh_error:
            mock_credentials = Mock()
            mock_credentials.refresh = Mock(side_effect=RefreshError)
            return mock_credentials, ""
        raise DefaultCredentialsError("abc")

    mock_run = Mock()
    mock_run.return_value.returncode = 0
    with patch(
        "anyscale.utils.gcp_utils.google.auth", default=Mock(side_effect=default)
    ), patch("anyscale.utils.gcp_utils.subprocess", run=mock_run):
        credentials, project = get_application_default_credentials(CloudSetupLogger())
        assert credentials is not None
        assert project == "project_abc"

    assert num_calls == 2
    mock_run.assert_called_once()
    _, err = capsys.readouterr()
    assert re.search(expected_output, err), err


@pytest.mark.parametrize(
    ("response", "valid_credentials"),
    [
        pytest.param(("200 OK", "service_get.json"), True, id="success"),
        pytest.param(("403 Forbidden", None), False, id="permission_denied"),
    ],
)
def test_get_google_cloud_client_factory(
    response, setup_mock_server, valid_credentials
):
    factory, tracker = setup_mock_server
    tracker.reset({".*": [response]})
    mock_project_id = "anyscale-bridge-deadbeef"
    if valid_credentials:
        with patch.multiple(
            "anyscale.utils.gcp_utils",
            get_application_default_credentials=Mock(
                return_value=(AnonymousCredentials(), mock_project_id)
            ),
            GoogleCloudClientFactory=Mock(return_value=factory),
        ):
            assert (
                get_google_cloud_client_factory(CloudSetupLogger(), mock_project_id)
                is not None
            )
    else:
        with patch.multiple(
            "anyscale.utils.gcp_utils",
            get_application_default_credentials=Mock(
                return_value=(AnonymousCredentials(), mock_project_id)
            ),
        ), pytest.raises(ClickException) as e:
            get_google_cloud_client_factory(CloudSetupLogger(), mock_project_id)
        e.match("Please make sure")


def test_get_filestore_location_and_instance_id():
    gcp_filestore_config = GCPFileStoreConfig(
        instance_name="projects/project_id/locations/fake_location/instances/fake_instance_id",
        root_dir="fake_root_dir",
        mount_target_ip="fake_mount_target_ip",
    )
    location, instance_id = get_filestore_location_and_instance_id(gcp_filestore_config)
    assert location == "fake_location"
    assert instance_id == "fake_instance_id"


def test_get_filestore_location_and_instance_id_exception():
    gcp_filestore_config = GCPFileStoreConfig(
        instance_name="projects/project_id",
        root_dir="fake_root_dir",
        mount_target_ip="fake_mount_target_ip",
    )
    with pytest.raises(ClickException) as e:
        get_filestore_location_and_instance_id(gcp_filestore_config)
    e.match("Could not parse Filestore instance name")



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class CloudAnalyticsEvent(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'trace_id': 'str',
        'cloud_id': 'str',
        'succeeded': 'bool',
        'command_name': 'CloudAnalyticsEventCommandName',
        'raw_command_input': 'str',
        'event_name': 'CloudAnalyticsEventName',
        'error': 'CloudAnalyticsEventError'
    }

    attribute_map = {
        'trace_id': 'trace_id',
        'cloud_id': 'cloud_id',
        'succeeded': 'succeeded',
        'command_name': 'command_name',
        'raw_command_input': 'raw_command_input',
        'event_name': 'event_name',
        'error': 'error'
    }

    def __init__(self, trace_id=None, cloud_id=None, succeeded=None, command_name=None, raw_command_input=None, event_name=None, error=None, local_vars_configuration=None):  # noqa: E501
        """CloudAnalyticsEvent - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._trace_id = None
        self._cloud_id = None
        self._succeeded = None
        self._command_name = None
        self._raw_command_input = None
        self._event_name = None
        self._error = None
        self.discriminator = None

        self.trace_id = trace_id
        if cloud_id is not None:
            self.cloud_id = cloud_id
        self.succeeded = succeeded
        self.command_name = command_name
        self.raw_command_input = raw_command_input
        self.event_name = event_name
        if error is not None:
            self.error = error

    @property
    def trace_id(self):
        """Gets the trace_id of this CloudAnalyticsEvent.  # noqa: E501

        The trace id generated by CLI  # noqa: E501

        :return: The trace_id of this CloudAnalyticsEvent.  # noqa: E501
        :rtype: str
        """
        return self._trace_id

    @trace_id.setter
    def trace_id(self, trace_id):
        """Sets the trace_id of this CloudAnalyticsEvent.

        The trace id generated by CLI  # noqa: E501

        :param trace_id: The trace_id of this CloudAnalyticsEvent.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and trace_id is None:  # noqa: E501
            raise ValueError("Invalid value for `trace_id`, must not be `None`")  # noqa: E501

        self._trace_id = trace_id

    @property
    def cloud_id(self):
        """Gets the cloud_id of this CloudAnalyticsEvent.  # noqa: E501

        The cloud id  # noqa: E501

        :return: The cloud_id of this CloudAnalyticsEvent.  # noqa: E501
        :rtype: str
        """
        return self._cloud_id

    @cloud_id.setter
    def cloud_id(self, cloud_id):
        """Sets the cloud_id of this CloudAnalyticsEvent.

        The cloud id  # noqa: E501

        :param cloud_id: The cloud_id of this CloudAnalyticsEvent.  # noqa: E501
        :type: str
        """

        self._cloud_id = cloud_id

    @property
    def succeeded(self):
        """Gets the succeeded of this CloudAnalyticsEvent.  # noqa: E501

        Whether the operation succeeded  # noqa: E501

        :return: The succeeded of this CloudAnalyticsEvent.  # noqa: E501
        :rtype: bool
        """
        return self._succeeded

    @succeeded.setter
    def succeeded(self, succeeded):
        """Sets the succeeded of this CloudAnalyticsEvent.

        Whether the operation succeeded  # noqa: E501

        :param succeeded: The succeeded of this CloudAnalyticsEvent.  # noqa: E501
        :type: bool
        """
        if self.local_vars_configuration.client_side_validation and succeeded is None:  # noqa: E501
            raise ValueError("Invalid value for `succeeded`, must not be `None`")  # noqa: E501

        self._succeeded = succeeded

    @property
    def command_name(self):
        """Gets the command_name of this CloudAnalyticsEvent.  # noqa: E501

        The command name (e.g. setup, register)  # noqa: E501

        :return: The command_name of this CloudAnalyticsEvent.  # noqa: E501
        :rtype: CloudAnalyticsEventCommandName
        """
        return self._command_name

    @command_name.setter
    def command_name(self, command_name):
        """Sets the command_name of this CloudAnalyticsEvent.

        The command name (e.g. setup, register)  # noqa: E501

        :param command_name: The command_name of this CloudAnalyticsEvent.  # noqa: E501
        :type: CloudAnalyticsEventCommandName
        """
        if self.local_vars_configuration.client_side_validation and command_name is None:  # noqa: E501
            raise ValueError("Invalid value for `command_name`, must not be `None`")  # noqa: E501

        self._command_name = command_name

    @property
    def raw_command_input(self):
        """Gets the raw_command_input of this CloudAnalyticsEvent.  # noqa: E501

        The command from user input (e.g. 'cloud verify my_cloud')  # noqa: E501

        :return: The raw_command_input of this CloudAnalyticsEvent.  # noqa: E501
        :rtype: str
        """
        return self._raw_command_input

    @raw_command_input.setter
    def raw_command_input(self, raw_command_input):
        """Sets the raw_command_input of this CloudAnalyticsEvent.

        The command from user input (e.g. 'cloud verify my_cloud')  # noqa: E501

        :param raw_command_input: The raw_command_input of this CloudAnalyticsEvent.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and raw_command_input is None:  # noqa: E501
            raise ValueError("Invalid value for `raw_command_input`, must not be `None`")  # noqa: E501

        self._raw_command_input = raw_command_input

    @property
    def event_name(self):
        """Gets the event_name of this CloudAnalyticsEvent.  # noqa: E501

        The event name (e.g. command_start, resources_created)  # noqa: E501

        :return: The event_name of this CloudAnalyticsEvent.  # noqa: E501
        :rtype: CloudAnalyticsEventName
        """
        return self._event_name

    @event_name.setter
    def event_name(self, event_name):
        """Sets the event_name of this CloudAnalyticsEvent.

        The event name (e.g. command_start, resources_created)  # noqa: E501

        :param event_name: The event_name of this CloudAnalyticsEvent.  # noqa: E501
        :type: CloudAnalyticsEventName
        """
        if self.local_vars_configuration.client_side_validation and event_name is None:  # noqa: E501
            raise ValueError("Invalid value for `event_name`, must not be `None`")  # noqa: E501

        self._event_name = event_name

    @property
    def error(self):
        """Gets the error of this CloudAnalyticsEvent.  # noqa: E501

        The error. It is only populated when succeeded is false.  # noqa: E501

        :return: The error of this CloudAnalyticsEvent.  # noqa: E501
        :rtype: CloudAnalyticsEventError
        """
        return self._error

    @error.setter
    def error(self, error):
        """Sets the error of this CloudAnalyticsEvent.

        The error. It is only populated when succeeded is false.  # noqa: E501

        :param error: The error of this CloudAnalyticsEvent.  # noqa: E501
        :type: CloudAnalyticsEventError
        """

        self._error = error

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, CloudAnalyticsEvent):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, CloudAnalyticsEvent):
            return True

        return self.to_dict() != other.to_dict()



from anyscale.client.openapi_client.models.apply_production_service_v2_model import (
    ApplyProductionServiceV2Model,
)
from anyscale.client.openapi_client.models.apply_service_config import (
    ApplyServiceConfig,
)
from anyscale.models.service_model import ServiceConfig


def format_service_config(service_config: ServiceConfig) -> ApplyServiceConfig:
    """
    This method formats the ServiceConfig into the API model ApplyServiceConfig.
    """
    service_name = service_config.name

    # We construct the Service v2 config model only if ray_serve_config is present
    if service_config.ray_serve_config:
        service_v2_config = ApplyProductionServiceV2Model(
            name=service_name,
            description=service_config.description or "Service updated from CLI",
            project_id=service_config.project_id,
            version=service_config.version,
            canary_percent=service_config.canary_percent,
            ray_serve_config=service_config.ray_serve_config,
            ray_gcs_external_storage_config=service_config.ray_gcs_external_storage_config,
            build_id=service_config.build_id,
            compute_config_id=service_config.compute_config_id,
        )
    else:
        service_v2_config = None
    return ApplyServiceConfig(service_v2_config=service_v2_config)


def format_service_config_v2(
    service_config: ServiceConfig,
) -> ApplyProductionServiceV2Model:
    if not service_config.ray_serve_config:
        raise RuntimeError(
            "ray_serve_config is expected for the Service configuration."
        )

    service_name = service_config.name

    return ApplyProductionServiceV2Model(
        name=service_name,
        description=service_config.description or "Service updated from CLI",
        project_id=service_config.project_id,
        version=service_config.version,
        canary_percent=service_config.canary_percent,
        ray_serve_config=service_config.ray_serve_config,
        ray_gcs_external_storage_config=service_config.ray_gcs_external_storage_config,
        build_id=service_config.build_id,
        compute_config_id=service_config.compute_config_id,
        rollout_strategy=service_config.rollout_strategy,
    )



from typing import Iterator, List
from unittest.mock import MagicMock, Mock, patch

from click import ClickException
import pytest

from anyscale.client.openapi_client.models import (
    ApplyProductionServiceV2Model,
    CloudProviders,
    ServiceConfig,
)
from anyscale.controllers.cloud_functional_verification_controller import (
    CloudFunctionalVerificationController,
    CloudFunctionalVerificationType,
    SERVICE_VERIFICATION_TIMEOUT_MINUTES,
)
from anyscale.sdk.anyscale_client.models import ComputeTemplateQuery


@pytest.fixture(autouse=True)
def mock_auth_api_client(base_mock_anyscale_api_client: Mock) -> Iterator[None]:
    mock_auth_api_client = Mock(
        api_client=Mock(), anyscale_api_client=base_mock_anyscale_api_client,
    )
    with patch.multiple(
        "anyscale.controllers.base_controller",
        get_auth_api_client=Mock(return_value=mock_auth_api_client),
    ):
        yield


@pytest.mark.parametrize("cloud_provider", [CloudProviders.AWS, CloudProviders.GCP])
@pytest.mark.parametrize("compute_config_exists", [True, False])
def test_get_or_create_cluster_compute(
    compute_config_exists: bool, cloud_provider: CloudProviders,
) -> None:
    mock_cloud_id = "mock_cloud_id"
    mock_cluster_compute_id = "mock_cluster_compute_id"
    mock_api_client = Mock()
    mock_api_client.search_compute_templates_api_v2_compute_templates_search_post = Mock(
        return_value=Mock(
            results=[Mock(id=mock_cluster_compute_id)] if compute_config_exists else []
        )
    )
    mock_anyscale_api_client = Mock()
    mock_anyscale_api_client.create_cluster_compute = Mock(
        return_value=Mock(result=Mock(id=mock_cluster_compute_id))
    )

    funciontal_verification_controller = CloudFunctionalVerificationController(Mock())
    funciontal_verification_controller.api_client = mock_api_client
    funciontal_verification_controller.anyscale_api_client = mock_anyscale_api_client
    assert (
        funciontal_verification_controller.get_or_create_cluster_compute(
            mock_cloud_id, cloud_provider
        )
        == mock_cluster_compute_id
    )

    mock_api_client.search_compute_templates_api_v2_compute_templates_search_post.assert_called_with(
        ComputeTemplateQuery(
            orgwide=True,
            name={"equals": f"functional_verification_{mock_cloud_id}"},
            include_anonymous=True,
        )
    )

    if compute_config_exists:
        mock_anyscale_api_client.create_cluster_compute.assert_not_called()
    else:
        mock_anyscale_api_client.create_cluster_compute.assert_called_once()


@pytest.mark.parametrize(
    ("prepare_verification_failed", "create_workspace_failed"),
    [
        pytest.param(False, False, id="happy-path"),
        pytest.param(True, False, id="prepare-verification-failed"),
        pytest.param(False, True, id="create-workspace-failed"),
    ],
)
def test_create_workspace(
    prepare_verification_failed: bool, create_workspace_failed: bool
):
    expected_result = not prepare_verification_failed and not create_workspace_failed
    mock_prepare_verification = Mock(return_value=(Mock(), Mock(), Mock()))
    if prepare_verification_failed:
        mock_prepare_verification.side_effect = ClickException("mock error")
    mock_create_workspace = Mock()
    if create_workspace_failed:
        mock_create_workspace.side_effect = ClickException("mock error")
    with patch.multiple(
        "anyscale.controllers.cloud_functional_verification_controller.CloudFunctionalVerificationController",
        _prepare_verification=mock_prepare_verification,
    ):
        controller = CloudFunctionalVerificationController(Mock())
        controller.api_client.create_workspace_api_v2_experimental_workspaces_post = (
            mock_create_workspace
        )
        if expected_result:
            controller.create_workspace(Mock(), Mock())
        else:
            with pytest.raises(ClickException):
                controller.create_workspace(Mock(), Mock())


@pytest.mark.parametrize(
    (
        "create_workspace_succeed",
        "poll_until_active_succeed",
        "terminate_workspace_succeed",
        "expected_result",
    ),
    [
        pytest.param(True, True, True, True, id="happy-path",),
        pytest.param(False, False, False, False, id="create-workspace-failed",),
        pytest.param(True, False, False, False, id="poll-until-active-failed",),
        pytest.param(True, True, False, False, id="workspace-termination-failed",),
    ],
)
def test_verify_workspace(
    create_workspace_succeed: bool,
    poll_until_active_succeed: bool,
    terminate_workspace_succeed: bool,
    expected_result: bool,
):
    mock_create_task = MagicMock()

    mock_terminate_workspace = (
        Mock()
        if terminate_workspace_succeed
        else Mock(side_effect=ClickException("mock error"))
    )
    with patch.multiple(
        "anyscale.controllers.cloud_functional_verification_controller.CloudFunctionalVerificationController",
        create_workspace=Mock()
        if create_workspace_succeed
        else Mock(side_effect=ClickException("mock error")),
        poll_until_active=Mock()
        if poll_until_active_succeed
        else Mock(side_effect=ClickException("mock error")),
        _create_task=mock_create_task,
    ):
        controller = CloudFunctionalVerificationController(Mock())
        controller.anyscale_api_client.terminate_cluster = mock_terminate_workspace
        assert controller.verify_workspace(Mock(), Mock()) == expected_result
        task_count = 1 + create_workspace_succeed + poll_until_active_succeed
        assert mock_create_task.call_count == task_count


@pytest.mark.parametrize("create_service_succeed", [True, False])
def test_rollout_service(create_service_succeed: bool):
    mock_cluster_compute_id = "mock_cluster_compute_id"
    mock_cluster_env_build_id = "mock_cluster_env_build_id"
    mock_project_id = "mock_project_id"
    mock_service_id = "mock_service_id"
    mock_cloud_id = "mock_cloud_id"
    mock_service_name = "mock_service_name"
    mock_version = "mock_version"
    mock_rollout_service = Mock(return_value=Mock(result=Mock(id=mock_service_id)))
    if not create_service_succeed:
        mock_rollout_service.side_effect = ClickException("mock error")
    with patch.multiple(
        "anyscale.controllers.cloud_functional_verification_controller.CloudFunctionalVerificationController",
        _prepare_verification=MagicMock(
            return_value=(
                mock_cluster_compute_id,
                mock_cluster_env_build_id,
                mock_project_id,
            )
        ),
    ):
        controller = CloudFunctionalVerificationController(Mock())
        controller.api_client.apply_service_v2_api_v2_services_v2_apply_put = (
            mock_rollout_service
        )

        if not create_service_succeed:
            with pytest.raises(ClickException):
                controller.rollout_service(
                    mock_cloud_id,
                    Mock(),
                    version=mock_version,
                    service_name=mock_service_name,
                )
        else:
            assert (
                controller.rollout_service(
                    mock_cloud_id,
                    Mock(),
                    version=mock_version,
                    service_name=mock_service_name,
                ).id
                == mock_service_id
            )

        mock_rollout_service.assert_called_once_with(
            ApplyProductionServiceV2Model(
                name=mock_service_name,
                description=f"service for cloud {mock_cloud_id} functional verification",
                project_id=mock_project_id,
                ray_serve_config={
                    "import_path": "serve_hello:entrypoint",
                    "runtime_env": {
                        "working_dir": "https://github.com/anyscale/docs_examples/archive/refs/heads/main.zip",
                        "env_vars": {
                            "SERVE_RESPONSE_MESSAGE": f"cloud functional verification {mock_version}",
                        },
                    },
                },
                version=mock_version,
                build_id=mock_cluster_env_build_id,
                compute_config_id=mock_cluster_compute_id,
                config=ServiceConfig(
                    max_uptime_timeout_sec=SERVICE_VERIFICATION_TIMEOUT_MINUTES * 60,
                ),
            )
        )


@pytest.mark.parametrize(
    (
        "deploy_service_succeed",
        "poll_until_active_succeed",
        "upgrade_service_succeed",
        "terminate_service_succeed",
    ),
    [
        pytest.param(True, True, True, True, id="happy-path"),
        pytest.param(False, False, False, False, id="deploy-service-failed"),
        pytest.param(True, False, False, False, id="poll-until-active-failed"),
        pytest.param(True, True, False, False, id="upgrade-service-failed"),
        pytest.param(True, True, True, False, id="terminate-service-failed"),
    ],
)
def test_verify_service(
    deploy_service_succeed: bool,
    poll_until_active_succeed: bool,
    upgrade_service_succeed: bool,
    terminate_service_succeed: bool,
):
    expected_result = (
        deploy_service_succeed
        and poll_until_active_succeed
        and upgrade_service_succeed
        and terminate_service_succeed
    )
    expected_task_num = (
        1 + deploy_service_succeed + poll_until_active_succeed + upgrade_service_succeed
    )

    mock_service_id = "mock_service_id"
    mock_service_name = "mock_service_name"

    mock_create_task = MagicMock()
    mock_rollout_service = Mock()
    if deploy_service_succeed and upgrade_service_succeed:
        mock_rollout_service = Mock(
            result=Mock(id=mock_service_id, name=mock_service_name)
        )
    elif deploy_service_succeed and not upgrade_service_succeed:
        mock_rollout_service.side_effect = [
            Mock(result=Mock(id=mock_service_id, name=mock_service_name)),
            ClickException("mock error"),
        ]
    else:
        mock_rollout_service.side_effect = ClickException("mock error")
    mock_terminate_service = Mock()
    if not terminate_service_succeed:
        mock_terminate_service.side_effect = ClickException("mock error")
    mock_poll_until_active = Mock()
    if not poll_until_active_succeed:
        mock_poll_until_active.side_effect = ClickException("mock error")

    with patch.multiple(
        "anyscale.controllers.cloud_functional_verification_controller.CloudFunctionalVerificationController",
        _create_task=mock_create_task,
        rollout_service=mock_rollout_service,
        poll_until_active=mock_poll_until_active,
    ):
        controller = CloudFunctionalVerificationController(Mock())
        controller.api_client.terminate_service_api_v2_services_v2_service_id_terminate_post = (
            mock_terminate_service
        )
        assert controller.verify_service(Mock(), Mock()) == expected_result
        assert mock_create_task.call_count == expected_task_num


@pytest.mark.parametrize(
    ("get_current_status_error", "status_not_allowed", "expected_result"),
    [
        pytest.param(False, False, True, id="happy-path"),
        pytest.param(True, False, False, id="get-current-status-error"),
        pytest.param(False, True, False, id="status-not-allowed"),
    ],
)
def test_poll_until_active(
    get_current_status_error, status_not_allowed, expected_result
):
    mock_function_type = Mock()
    mock_function_id = "mock_function_id"
    mock_cluster_id = "mock_cluster_id"
    mock_function = Mock(id=mock_function_id, cluster_id=mock_cluster_id)
    goal_status = "goal_status"
    mock_get_current_status = Mock(
        return_value="not_allowed" if status_not_allowed else goal_status
    )
    if get_current_status_error:
        mock_get_current_status.side_effect = ClickException("mock error")
    with patch.multiple(
        "anyscale.controllers.cloud_functional_verification_controller.CloudFunctionalVerificationController",
        _update_task_in_step_progress=Mock(),
        _render_event_log=Mock(),
    ), patch.multiple(
        "anyscale.controllers.cloud_functional_verification_controller",
        POLL_INTERVAL_SECONDS=0,
    ):
        controller = CloudFunctionalVerificationController(Mock())
        if expected_result:
            assert (
                controller.poll_until_active(
                    mock_function_type,
                    mock_function,
                    mock_get_current_status,
                    goal_status,
                    set(goal_status),
                    Mock(),
                    1,
                )
                == expected_result
            )
            mock_get_current_status.assert_called_once_with(mock_function_id)
        else:
            with pytest.raises(ClickException) as e:
                controller.poll_until_active(
                    mock_function_type,
                    mock_function,
                    mock_get_current_status,
                    goal_status,
                    set(goal_status),
                    Mock(),
                    1,
                )
                if get_current_status_error:
                    assert e.match("Failed to get")
                if status_not_allowed:
                    assert e.match("is in an unexpected state:")


def test_poll_until_active_timeout():
    mock_function_type = Mock()
    mock_function_id = "mock_function_id"
    mock_cluster_id = "mock_cluster_id"
    mock_function = Mock(id=mock_function_id, cluster_id=mock_cluster_id)
    goal_status = "goal_status"
    mock_status = "mock_status"
    allowed_status_set = {goal_status, mock_status}
    mock_get_current_status = Mock(return_value=mock_status)
    with patch.multiple(
        "anyscale.controllers.cloud_functional_verification_controller.CloudFunctionalVerificationController",
        _update_task_in_step_progress=Mock(),
    ), patch.multiple(
        "anyscale.controllers.cloud_functional_verification_controller",
        POLL_INTERVAL_SECONDS=0,
    ), pytest.raises(
        ClickException
    ) as e:
        controller = CloudFunctionalVerificationController(Mock())
        assert (
            controller.poll_until_active(
                mock_function_type,
                mock_function,
                mock_get_current_status,
                goal_status,
                allowed_status_set,
                Mock(),
                0,
            )
            is False
        )
        assert e.match("Timed out")


@pytest.mark.parametrize("verification_result", [True, False])
@pytest.mark.parametrize("cloud_provider", [CloudProviders.AWS, CloudProviders.GCP])
@pytest.mark.parametrize(
    "functions_to_verify",
    [
        [CloudFunctionalVerificationType.WORKSPACE],
        [
            CloudFunctionalVerificationType.WORKSPACE,
            CloudFunctionalVerificationType.SERVICE,
        ],
    ],
)
def test_start_verification(
    functions_to_verify: List[CloudFunctionalVerificationType],
    cloud_provider: CloudProviders,
    verification_result: bool,
):
    mock_cloud_id = "mock_cloud_id"

    with patch.multiple(
        "anyscale.controllers.cloud_functional_verification_controller.CloudFunctionalVerificationController",
        verify=Mock(return_value=verification_result),
        get_live_console=MagicMock(),
        _prepare_verification=Mock(),
    ):
        funciontal_verification_controller = CloudFunctionalVerificationController(
            Mock()
        )
        assert (
            funciontal_verification_controller.start_verification(
                mock_cloud_id, cloud_provider, functions_to_verify, yes=True
            )
            == verification_result
        )


@pytest.mark.parametrize(
    "functions_to_verify",
    [
        [CloudFunctionalVerificationType.WORKSPACE],
        [
            CloudFunctionalVerificationType.WORKSPACE,
            CloudFunctionalVerificationType.SERVICE,
        ],
    ],
)
def test_get_live_console(functions_to_verify):
    funciontal_verification_controller = CloudFunctionalVerificationController(Mock())
    funciontal_verification_controller.get_live_console(functions_to_verify)
    for function in functions_to_verify:
        assert funciontal_verification_controller.step_progress[function] is not None
        assert funciontal_verification_controller.overall_progress[function] is not None
        assert funciontal_verification_controller.task_ids[function] is not None



from dataclasses import dataclass
import inspect
from typing import Any, Callable, Dict, Generic, List, Optional, TypeVar

from pydantic import BaseModel

from anyscale.client.openapi_client.models.session import Session
from anyscale.sdk.anyscale_client.models.cluster import Cluster
from anyscale.sdk.anyscale_client.sdk import AnyscaleSDK


try:
    from ray.client_builder import ClientContext
except ImportError:
    # Import error will be raised in `ClientBuilder.__init__`.
    # Not raising error here to allow `anyscale.connect required_ray_version`
    # to work.
    @dataclass
    class ClientContext:  # type: ignore
        pass


@dataclass
class AnyscaleClientConnectResponse:
    """
    Additional information returned about clusters that were connected from Anyscale.
    """

    cluster_id: str


class AnyscaleClientContext(ClientContext):  # type: ignore
    def __init__(
        self, anyscale_cluster_info: AnyscaleClientConnectResponse, **kwargs: Any,
    ) -> None:
        if _multiclient_supported() and "_context_to_restore" not in kwargs:
            # Set to None for now until multiclient is supported on connect
            kwargs["_context_to_restore"] = None
        super().__init__(**kwargs)
        self.anyscale_cluster_info = anyscale_cluster_info


def _multiclient_supported() -> bool:
    """True if ray version supports multiple clients, False otherwise"""
    _context_params = inspect.signature(ClientContext.__init__).parameters
    return "_context_to_restore" in _context_params


def find_project_id(sdk: AnyscaleSDK, project_name: str) -> Optional[str]:
    """Return id if a project of a given name exists."""
    resp = sdk.search_projects({"name": {"equals": project_name}})
    if len(resp.results) > 0:
        return resp.results[0].id  # type: ignore
    else:
        return None


def get_cluster(
    sdk: AnyscaleSDK, project_id: str, session_name: str
) -> Optional[Session]:
    """Query Anyscale for the given cluster's metadata."""
    clusters = sdk.search_clusters(
        {
            "project_id": project_id,
            "name": {"equals": session_name},
            "archive_status": "ALL",
        }
    ).results
    cluster_found: Optional[Cluster] = None
    for cluster in clusters:
        if cluster.name == session_name:
            cluster_found = cluster
            break

    # A separate call is needed because downstream clients expect "Session" type
    if cluster_found:
        return sdk.get_session(cluster_found.id).result
    return None


def list_entities(
    list_function: Callable[..., Any],
    container_id: Optional[str] = None,
    filters: Optional[Dict[str, Any]] = None,
    max: Optional[int] = None,  # noqa: A002
) -> List[Any]:
    """Convenience function to automatically handle paging tokens.

    This function repeatedly calls `list_function`, which is a GET endpoint
    with `container_id` until all (potentially paged) results are received.
    """
    entities: List[Any] = []
    has_more = True
    paging_token = None
    filters = filters or {}
    while has_more and (not max or len(entities) < max):
        if container_id:
            resp = list_function(
                container_id, count=50, paging_token=paging_token, **filters
            )
        else:
            resp = list_function(count=50, paging_token=paging_token, **filters)
        entities.extend(resp.results)
        paging_token = resp.metadata.next_paging_token
        has_more = paging_token is not None
    return entities


T = TypeVar("T")


class ListResponseMetadata(BaseModel):
    total: Optional[int] = None
    next_paging_token: Optional[str] = None


class ListResponse(BaseModel, Generic[T]):
    results: List[T]
    metadata: ListResponseMetadata


# Must be the `<Entity>Query` Anyscale model. It cannot be a dict.
QueryObj = TypeVar("QueryObj", bound=object)


def search_entities(
    search_function: Callable[[QueryObj], ListResponse[T]],
    query_obj: QueryObj,
    max_to_return: int = 1000,
) -> List[T]:
    """Automatically paginate to retrieve all search results up to `max_to_return`
    (default = 1000).

    Works for POST endpoints where a single query object is passed
    (in the request body) to the endpoint.
    """
    entities: List[T] = []
    has_more = True
    count = 0

    while has_more and count < max_to_return:
        resp = search_function(query_obj)
        count += len(resp.results)
        entities += resp.results

        next_paging_token = resp.metadata.next_paging_token
        query_obj.paging.paging_token = next_paging_token  # type: ignore
        has_more = next_paging_token is not None
    if count > max_to_return:
        count = max_to_return
        entities = entities[:max_to_return]

    return entities
